"""
Bumps command line interface.

The functions in this module are used by the bumps command to implement
the command line interface.  Bumps plugin models can use them to create
stand alone applications with a similar interface.  For example, the
Refl1D application uses the following::

    from . import fitplugin
    import bumps.cli
    bumps.cli.set_mplconfig(appdatadir='Refl1D')
    bumps.cli.install_plugin(fitplugin)
    bumps.cli.main()

After completing a set of fits on related systems, a post-analysis script
can use :func:`load_model` to load the problem definition and
:func:`load_best` to load the best value  found in the fit.  This can
be used for example in experiment design, where you look at the expected
parameter uncertainty when fitting simulated data from a range of experimental
systems.
"""

__all__ = [
    "main",
    "install_plugin",
    "set_mplconfig",
    "config_matplotlib",
    "load_model",
    "preview",
    "load_best",
    "save_best",
    "resynth",
]

import sys
import os
import re
import warnings
import traceback

import shutil

import numpy as np
# np.seterr(all="raise")

from . import fitters
from .fitters import FitDriver, StepMonitor, ConsoleMonitor, CheckpointMonitor, nllf_scale
from .mapper import MPMapper, AMQPMapper, MPIMapper, SerialMapper
from .formatnum import format_uncertainty
from . import util
from . import initpop
from . import __version__
from . import plugin
from . import options

from .util import pushdir, push_python_path


def install_plugin(p):
    """
    Replace symbols in :mod:`bumps.plugin` with application specific
    methods.
    """
    for symbol in plugin.__all__:
        if hasattr(p, symbol):
            setattr(plugin, symbol, getattr(p, symbol))


def load_model(path, model_options=None):
    """
    Load a model file.

    *path* contains the path to the model file.

    *model_options* are any additional arguments to the model.  The sys.argv
    variable will be set such that *sys.argv[1:] == model_options*.
    """
    from .fitproblem import load_problem

    # Change to the target path before loading model so that data files
    # can be given as relative paths in the model file.  Add the directory
    # to the python path (at the end) so that imports work as expected.
    directory, filename = os.path.split(path)
    with pushdir(directory):
        # Try a specialized model loader
        problem = plugin.load_model(filename)
        if problem is None:
            # print "loading",filename,"from",directory
            # TODO: eliminate pickle!!
            if filename.endswith("pickle"):
                try:
                    import dill as pickle
                except ImportError:
                    import pickle
                # First see if it is a pickle
                with open(filename, "rb") as fd:
                    problem = pickle.load(fd)
            else:
                # Then see if it is a python model script
                problem = load_problem(filename, options=model_options)

    # Guard against the user changing parameters after defining the problem.
    problem.model_reset()
    problem.path = os.path.abspath(path)
    if not hasattr(problem, "title"):
        problem.title = filename
    problem.name, _ = os.path.splitext(filename)
    problem.options = model_options
    return problem


def preview(problem, view=None):
    """
    Show the problem plots and parameters.
    """
    import pylab

    problem.show()
    problem.plot(view=view)
    pylab.show()


def save_best(fitdriver, problem, best, view=None):
    """
    Save the fit data, including parameter values, uncertainties and plots.

    *fitdriver* is the fitter that was used to drive the fit.

    *problem* is a FitProblem instance.

    *best* is the parameter set to save.
    """
    # Make sure the problem contains the best value
    # TODO: avoid recalculating if problem is already at best.
    problem.setp(best)
    # print "remembering best"
    pardata = "".join("%s %.15g\n" % (name, value) for name, value in zip(problem.labels(), problem.getp()))
    open(problem.output_path + ".par", "wt").write(pardata)

    fitdriver.save(problem.output_path)
    with util.redirect_console(problem.output_path + ".err"):
        fitdriver.show()
        fitdriver.plot(output_path=problem.output_path, view=view)
    fitdriver.show()
    # print "plotting"


PARS_PATTERN = re.compile(r"^(?P<label>.*) (?P<value>[^ ]*)\n$")


def load_best(problem, path):
    """
    Reload individual parameter values from a saved .par file.

    If the label does not exist in the file, use the value from the model
    as the default value. Ignore labels that do not exist in the model. In
    that way we can load parameters from an old fit with minimal fuss, even
    as we add, delete and move parameters in the model. If any parameters
    are missing, set *problem.undefined* to the a boolean index of the
    undefined parameters.

    There is an interaction with --init=eps and the par file. If any parameters
    are missing from the par file they will be randomized across the
    entire parameter range using the equivalent of --init=lhs. That means
    you can drop a # at the beginning of the line in the .par file
    and that parameter will be shuffled on restart, with the remaining
    parameters starting near the initial value.
    """
    # WARNING: Labels are not unique! Need to track multiple instances of
    # the same label.
    if not os.path.isfile(path):
        path = os.path.join(path, problem.name + ".par")
    if not os.path.isfile(path):
        raise ValueError("Parameter file %s does not exist." % path)
    labels = problem.labels()
    targets = {label: [] for label in labels}
    with open(path, "rt") as fid:
        for line in fid:
            m = PARS_PATTERN.match(line)
            label, value = m.group("label"), float(m.group("value"))
            # Accumulate values for labels only if they appear in the model.
            if label in targets:
                targets[label].append(value)

    # Populate model with named parameters in the order they occur in the
    # parameter file. Identify the missing parameters if any, adding them
    # to the the problem definition as an optional "undefined" attribute with
    # one bit for each parameter. This ugly hack is to support a previous
    # ugly hack in which undefined parameters are initialized with LHS but
    # defined parameters are initialized with eps, cov or random.
    # TODO: find a better way to "free" parameters on --pars.
    # TODO: find a way to "free" parameters on --resume.
    values, undefined = [], []
    for label, default_value in zip(labels, problem.getp()):
        remaining = targets[label]
        is_empty = not remaining
        # popping the next value from remaining modifies targets[label]
        values.append(default_value if is_empty else remaining.pop(0))
        undefined.append(is_empty)
    problem.setp(np.asarray(values))
    if any(undefined):
        problem.undefined = np.asarray(undefined)


# CRUFT
recall_best = load_best


def store_overwrite_query_gui(path):
    """
    Ask if store path should be overwritten.

    Use this in a call to :func:`make_store` from a graphical user interface.
    """
    import wx

    msg = path + " already exists. Press 'yes' to overwrite, or 'No' to abort and restart with newpath"
    msg_dlg = wx.MessageDialog(None, msg, "Overwrite Directory", wx.YES_NO | wx.ICON_QUESTION)
    retCode = msg_dlg.ShowModal()
    msg_dlg.Destroy()
    if retCode != wx.ID_YES:
        raise RuntimeError("Could not create path")


def store_overwrite_query(path):
    """
    Ask if store path should be overwritten.

    Use this in a call to :func:`make_store` from a command line interface.
    """
    print(path, "already exists.")
    print("Press 'y' to overwrite, or 'n' to abort and restart with --overwrite, --resume, or --store=newpath")
    ans = input("Overwrite [y/n]? ")
    if ans not in ("y", "Y", "yes"):
        sys.exit(1)


def make_store(problem, opts, exists_handler):
    """
    Create the store directory and populate it with the model definition file.
    """
    # Determine if command line override
    if opts.store:
        problem.store = opts.store
    if getattr(problem, "store", None) is None:
        raise RuntimeError("Need to specify '--store=path' on command line or problem.store='path' in definition file.")
    problem.output_path = os.path.join(problem.store, problem.name)

    # Check if already exists
    store_exists = os.path.exists(problem.output_path + ".par")
    if not opts.overwrite and opts.resume is None and store_exists:
        if opts.batch:
            print(
                problem.store + " already exists.  Restart with --overwrite, --resume, or --store=newpath",
                file=sys.stderr,
            )
            sys.exit(1)
        exists_handler(problem.output_path)

    # Create it and copy model
    if not os.path.exists(problem.store):
        os.mkdir(problem.store)
    shutil.copy2(problem.path, problem.store)


def run_profiler(problem, steps):
    """
    Model execution profiler.

    Run the program with "--profiler --steps=N" to generate a function
    profile chart breaking down the cost of evaluating N models.
    """
    # Here is the findings from one profiling session::
    #   23 ms total
    #    6 ms rendering model
    #    8 ms abeles
    #    4 ms convolution
    #    1 ms setting parameters and computing nllf
    from .util import profile

    p = initpop.random_init(int(steps), initial=None, bounds=None, use_point=False, problem=problem)
    # Note: map is an iterator in python 3
    profile(lambda *args: list(map(*args)), problem.nllf, p)


def run_timer(mapper, problem, steps):
    """
    Model execution timer.

    Run the program with "--timer --steps=N" to determine the average
    run time of the model.  If --parallel is included, then the model
    will be run in parallel on separate cores.
    """
    import time

    T0 = time.time()
    steps = int(steps)
    p = initpop.generate(problem, init="random", pop=-steps, use_point=False)
    if p.shape == (0,):
        # No fitting parameters --- generate an empty population
        p = np.empty((steps, 0))
    mapper(p)
    print("time per model eval: %g ms" % (1000 * (time.time() - T0) / steps,))


def start_remote_fit(problem, options, queue, notify):
    """
    Queue remote fit.
    """
    from jobqueue.client import connect

    try:
        from dill import dumps as dill_dumps

        dumps = lambda obj: dill_dumps(obj, recurse=True)
    except ImportError:
        from pickle import dumps

    data = dict(package="bumps", version=__version__, problem=dumps(problem), options=dumps(options))
    request = dict(
        service="fitter",
        version=__version__,  # fitter service version
        notify=notify,
        name=problem.title,
        data=data,
    )

    server = connect(queue)
    job = server.submit(request)
    return job


# ==== Main ====


def initial_model(opts):
    """
    Load and initialize the model.

    *opts* are the processed command line options.

    If --pars is in opts, then load the parameters from a .par file.

    If --simulate is in opts, then generate random data from the model.

    If --simrandom is in opts, then generate random data from a random model.

    If --shake is in opts, then use a random initial state for the fit.
    """
    if opts.seed is not None:
        np.random.seed(opts.seed)

    if opts.args:
        problem = load_model(opts.args[0], opts.args[1:])
        if opts.pars is not None:
            load_best(problem, opts.pars)
        if opts.simrandom:
            problem.randomize()
        if opts.simulate or opts.simrandom:
            noise = None if opts.noise == "data" else float(opts.noise)
            problem.simulate_data(noise=noise)
            print("simulation parameters")
            print(problem.summarize())
            print("chisq at simulation", problem.chisq_str())
        if opts.shake:
            problem.randomize()
    else:
        problem = None
    return problem


def resynth(fitdriver, problem, mapper, opts):
    """
    Generate maximum likelihood fits to resynthesized data sets.

    *fitdriver* is a :class:`bumps.fitters.FitDriver` object with a fitter
    already chosen.

    *problem* is a :func:`bumps.fitproblem.FitProblem` object.  It should
    be initialized with optimal values for the parameters.

    *mapper* is one of the available :mod:`bumps.mapper` classes.

    *opts* is a :class:`bumps.options.BumpsOpts` object representing the command
    line parameters.
    """
    make_store(problem, opts, exists_handler=store_overwrite_query)
    fid = open(problem.output_path + ".rsy", "at")
    fitdriver.mapper = mapper.start_mapper(problem, opts.args)
    for i in range(opts.resynth):
        problem.resynth_data()
        best, fbest = fitdriver.fit()
        scale, err = nllf_scale(problem)
        print("step %d chisq %g" % (i, scale * fbest))
        fid.write("%.15g " % (scale * fbest))
        fid.write(" ".join("%.15g" % v for v in best))
        fid.write("\n")
    problem.restore_data()
    fid.close()


def set_mplconfig(appdatadir):
    r"""
    Point the matplotlib config dir to %LOCALAPPDATA%\{appdatadir}\mplconfig.
    """
    if hasattr(sys, "frozen"):
        if os.name == "nt":
            mplconfigdir = os.path.join(os.environ["LOCALAPPDATA"], appdatadir, "mplconfig")
        elif sys.platform == "darwin":
            mplconfigdir = os.path.join(os.path.expanduser("~/Library/Caches"), appdatadir, "mplconfig")
        else:
            return  # do nothing on linux
        mplconfigdir = os.environ.setdefault("MPLCONFIGDIR", mplconfigdir)
        if not os.path.exists(mplconfigdir):
            os.makedirs(mplconfigdir)


def config_matplotlib(backend=None):
    """
    Setup matplotlib to use a particular backend.

    The backend should be 'WXAgg' for interactive use, or 'Agg' for batch.
    This distinction allows us to run in environments such as cluster computers
    which do not have wx installed on the compute nodes.

    This function must be called before any imports to pylab.  To allow
    this, modules should not import pylab at the module level, but instead
    import it for each function/method that uses it.  Exceptions can be made
    for modules which are completely dedicated to plotting, but these modules
    should never be imported at the module level.
    """
    import matplotlib as mpl

    # When running from a frozen environment created by py2exe, we will not
    # have a range of backends available, and must set the default to WXAgg.
    # With a full matplotlib distribution we can use whatever the user prefers.
    if hasattr(sys, "frozen"):
        if "MPLCONFIGDIR" not in os.environ:
            raise RuntimeError(r"MPLCONFIGDIR should be set to e.g., %LOCALAPPDATA%\YourApp\mplconfig")
        if backend is None:
            backend = "WXAgg"

    ## CRUFT: check that backend is valid, trying alternates if an import fails
    # if backend is None:
    #    backend = os.environ.get('MPLBACKEND', mpl.rcParams['backend'])
    # import importlib
    # for name in (backend, 'MacOSX', 'Qt5Agg', 'Qt4Agg', 'Gtk3Agg', 'TkAgg', 'WXAgg'):
    #    path = 'matplotlib.backends.backend_' + name.lower()
    #    try:
    #        importlib.import_module(path)
    #        backend = name
    #        break
    #    except ImportError:
    #        backend = None

    # Specify the backend to use for plotting and import backend dependent
    # classes.  This must be done before importing pyplot to have an
    # effect.  If no backend is given, let pyplot use the default.
    if backend is not None:
        mpl.use(backend)

    # Disable interactive mode so that plots are only updated on show() or
    # draw(). The interactive function must be called before importing pyplot,
    # otherwise it will have no effect.
    mpl.interactive(False)

    # configure the plot style
    line_width = 1
    pad = 2
    font_family = "Arial" if os.name == "nt" else "sans-serif"
    font_size = 12
    plot_style = {
        "xtick.direction": "in",
        "ytick.direction": "in",
        "lines.linewidth": line_width,
        "axes.linewidth": line_width,
        "xtick.labelsize": font_size,
        "ytick.labelsize": font_size,
        "xtick.major.size": 5,
        "ytick.major.size": 5,
        "xtick.minor.size": 2.5,
        "ytick.minor.size": 2.5,
        "xtick.major.width": line_width,
        "ytick.major.width": line_width,
        "xtick.minor.width": line_width,
        "ytick.minor.width": line_width,
        "xtick.major.pad": pad,
        "ytick.major.pad": pad,
        "xtick.top": True,
        "ytick.right": True,
        "font.size": font_size,
        "font.family": font_family,
        "svg.fonttype": "none",
        "savefig.dpi": 100,
    }
    mpl.rcParams.update(plot_style)


def beep():
    """
    Audio signal that fit is complete.
    """
    if sys.platform == "win32":
        try:
            import winsound

            winsound.MessageBeep(winsound.MB_OK)
        except Exception:
            pass
    else:
        print("\a", file=sys.__stdout__)


def run_command(c):
    """
    Run an arbitrary python command.
    """
    exec(c, globals())


def setup_logging():
    """Start logger"""
    import logging

    logging.basicConfig(level=logging.INFO)


# From http://stackoverflow.com/questions/22373927/get-traceback-of-warnings
# answered by mgab (2014-03-13)
# edited by Gareth Rees (2015-11-28)
def warn_with_traceback(message, category, filename, lineno, file=None, line=None):
    """
    Alternate warning printer which shows a traceback with the warning.

    To use, set *warnings.showwarning = warn_with_traceback*.
    """
    traceback.print_stack()
    log = file if hasattr(file, "write") else sys.stderr
    log.write(warnings.formatwarning(message, category, filename, lineno, line))


def main():
    """
    Run the bumps program with the command line interface.

    Input parameters are taken from sys.argv.
    """
    # add full traceback to warnings
    # warnings.showwarning = warn_with_traceback

    if len(sys.argv) == 1:
        sys.argv.append("-?")

    # run command with bumps in the environment
    if sys.argv[1] == "-m":
        import runpy

        sys.argv = sys.argv[2:]
        runpy.run_module(sys.argv[0], run_name="__main__")
        sys.exit(0)
    elif sys.argv[1] == "-p":
        import runpy

        sys.argv = sys.argv[2:]
        runpy.run_path(sys.argv[0], run_name="__main__")
        sys.exit()
    elif sys.argv[1] == "-c":
        run_command(sys.argv[2])
        sys.exit()
    elif sys.argv[1] == "-i":
        sys.argv = ["ipython", "--pylab"]
        from IPython import start_ipython

        sys.exit(start_ipython())

    opts = options.getopts()
    setup_logging()

    if opts.edit:
        from .gui.gui_app import main as gui

        gui()
        return

    # Set up the matplotlib backend to minimize the wx/gui dependency.
    # If no GUI specified and not editing, then use the default mpl
    # backend for the python version.
    if opts.batch or opts.remote or opts.noshow:  # no interactivity
        config_matplotlib(backend="agg")
    else:  # let preview use default graphs
        config_matplotlib()

    problem = initial_model(opts)
    if problem is None:
        print("\n!!! Model file missing from command line --- abort !!!.", file=sys.stderr)
        sys.exit(1)

    # TODO: AMQP mapper as implemented requires workers started up with
    # the particular problem; need to be able to transport the problem
    # to the worker instead.  Until that happens, the GUI shouldn't use
    # the AMQP mapper.
    if opts.mpi:
        MPIMapper.start_worker(problem)
        mapper = MPIMapper
    elif opts.parallel != "" or opts.worker:
        if opts.transport == "amqp":
            mapper = AMQPMapper
        elif opts.transport == "mp":
            mapper = MPMapper
        elif opts.transport == "celery":
            mapper = CeleryMapper
        else:
            raise ValueError("unknown mapper")
    else:
        mapper = SerialMapper
    if opts.worker:
        mapper.start_worker(problem)
        return

    if np.isfinite(float(opts.time)):
        import time

        start_time = time.time()
        stop_time = start_time + float(opts.time) * 3600
        abort_test = lambda: time.time() >= stop_time
    else:
        abort_test = lambda: False

    fitdriver = FitDriver(
        opts.fit_config.selected_fitter, problem=problem, abort_test=abort_test, **opts.fit_config.selected_values
    )

    # Start fitter within the domain so that constraints are valid
    clipped = fitdriver.clip()
    if clipped:
        print("Start value clipped to range for parameter", ", ".join(clipped))

    if opts.time_model:
        run_timer(mapper.start_mapper(problem, opts.args), problem, steps=int(opts.steps))
    elif opts.profile:
        run_profiler(problem, steps=int(opts.steps))
    elif opts.chisq:
        if opts.cov:
            fitdriver.show_cov()
        print("chisq", problem.chisq_str())
        # import pprint; pprint.pprint(problem.to_dict(), indent=2, width=272)
    elif opts.preview:
        if opts.cov:
            fitdriver.show_cov()
        preview(problem, view=opts.view)
    elif opts.resynth > 0:
        resynth(fitdriver, problem, mapper, opts)

    elif opts.remote:
        # Check that problem runs before submitting it remotely
        # TODO: this may fail if problem requires remote resources such as GPU
        print("initial chisq:", problem.chisq_str())
        job = start_remote_fit(problem, opts, queue=opts.queue, notify=opts.notify)
        print("remote job:", job["id"])

    else:
        # Show command line arguments and initial model
        print("#", " ".join(sys.argv), "--seed=%d" % opts.seed)
        problem.show()

        # Check that there are parameters to be fitted.
        if not len(problem.getp()):
            print("\n!!! No parameters selected for fitting---abort !!!\n", file=sys.stderr)
            sys.exit(1)

        # Run the fit
        if opts.resume == "-":
            opts.resume = opts.store if os.path.exists(opts.store) else None
        if opts.resume:
            resume_path = os.path.join(opts.resume, problem.name)
        else:
            resume_path = None

        make_store(problem, opts, exists_handler=store_overwrite_query)

        # Redirect sys.stdout to capture progress
        if opts.batch:
            sys.stdout = open(problem.output_path + ".mon", "w")

        # TODO: fix techical debt with checkpoint monitor implementation
        # * The current checkpoint implementation is self-referential:
        #     checkpoint = lambda: save_best(fitdriver, ...)
        #     fitdriver.monitors = [..., CheckpointMonitor(checkpoint), ...]
        #   It is done this way because the checkpoint monitor needs the fitter
        #   so it can ask it to save state, but the fitter needs the list of
        #   monitors, including the checkpoint monitor, before it is run.
        # * Figures are cumulative, with each checkpoint adding a new set
        # * Figures are slow! Can they go into a separate thread?  Can we
        #   have the problem cache the best value?
        checkpoint_time = float(opts.checkpoint) * 3600

        def checkpoint(history):
            problem = fitdriver.problem
            ## Use the following to save only the fitter state
            fitdriver.fitter.save(problem.output_path)
            ## Use the following to save the fitter state plus all other
            ## plots and other output files.  This won't work yet since
            ## plots are generated sequentially, with each checkpoint producing
            ## a completely new set of plots.
            # best = history.point[0]
            # save_best(fitdriver, problem, best, view=opts.view)

        monitors = [ConsoleMonitor(problem)]
        if checkpoint_time > 0 and np.isfinite(checkpoint_time):
            mon = CheckpointMonitor(checkpoint, progress=checkpoint_time)
            monitors.append(mon)
        if opts.stepmon:
            fid = open(problem.output_path + ".log", "w")
            mon = StepMonitor(problem, fid, fields=["step", "value"])
            monitors.append(mon)
        fitdriver.monitors = monitors

        # import time; t0=time.clock()
        cpus = int(opts.parallel) if opts.parallel != "" else 0
        fitdriver.mapper = mapper.start_mapper(problem, opts.args, cpus=cpus)
        best, fbest = fitdriver.fit(resume=resume_path)
        # print("time=%g"%(time.clock()-t0),file=sys.__stdout__)
        # Note: keep this in sync with the checkpoint function above
        save_best(fitdriver, problem, best, view=opts.view)
        if opts.err or opts.cov:
            fitdriver.show_err()
        if opts.cov:
            fitdriver.show_cov()
        if opts.entropy:
            fitdriver.show_entropy(opts.entropy)
        mapper.stop_mapper(fitdriver.mapper)

        # If in batch mode then explicitly close the monitor file on completion
        if opts.batch:
            sys.stdout.close()
            sys.stdout = sys.__stdout__

        # Display the plots
        if not opts.batch and not opts.mpi and not opts.noshow:
            beep()
            import pylab

            pylab.show()


# Allow  "$python -m bumps.cli args" calling pattern
if __name__ == "__main__":
    main()



"""
Convergence test statistic from Gelman and Rubin, 1992.
"""

__all__ = ["geweke"]

from numpy import var, mean, ones, sqrt, reshape, log10, abs


def geweke(sequences, portion=0.25):
    """
    Calculates the Geweke convergence diagnostic

    Refer to:

        pymc-devs.github.com/pymc/modelchecking.html#informal-methods
        support.sas.com/documentation/cdl/en/statug/63033/HTML/default/viewer.htm#statug_introbayes_sect008.html

    """

    # Find the size of the sample
    chain_len, nchains, nvar = sequences.shape
    z_stat = -2 * ones(nvar)
    if chain_len >= 2:
        # Only use the last portion of the sample
        try:
            front_portion, back_portion = portion
        except TypeError:
            front_portion = back_portion = portion
        front_len = int(chain_len * front_portion)
        back_len = int(chain_len * back_portion)
        # print "STARTING SHAPE", sequences.shape
        seq1 = reshape(sequences[:front_len, :, :], (front_len * nchains, nvar))
        seq2 = reshape(sequences[-back_len:, :, :], (back_len * nchains, nvar))
        # print "SEQ1", seq1.shape, 'SEQ2', seq2.shape
        # Step 1: Determine the sequence means
        meanseq1 = mean(seq1, axis=0)
        meanseq2 = mean(seq2, axis=0)
        # print "SHAPEs", meanseq1.shape, meanseq2.shape
        var1 = var(seq1, axis=0)
        var2 = var(seq2, axis=0)
        denom = sqrt(var1 + var2)
        index = denom > 0
        z_stat[index] = (meanseq1 - meanseq2)[index] / denom[index]

        # z_stat is now the Z score for every chain and parameter
        # in that with shape (chains, vars)

        # To make it easier to look at, return the average for the vars.
        if 0:
            avg_z = mean(z_stat, axis=0)
            lavg_z = log10(abs(avg_z))
            return lavg_z.tolist()
        if 0:
            avg_z = z_stat
            lavg_z = log10(abs(avg_z))
            return lavg_z.flatten().tolist()
        else:
            return z_stat.flatten().tolist()

    # TODO: code is wrong if chain length is 1, since lavg_z is not defined
    return lavg_z.tolist()



import logging


class ListHandler(logging.Handler):
    def __init__(self):
        super().__init__()
        self.log_list = []

    def emit(self, record):
        self.log_list.append(self.format(record))


# Create a logger
logger = logging.getLogger("webview")
logger.setLevel(logging.INFO)

# Create and add the custom handler
list_handler = ListHandler()
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
list_handler.setFormatter(formatter)
list_handler.setLevel(logging.INFO)

console_handler = logging.StreamHandler()
console_handler.setFormatter(formatter)
console_handler.setLevel(logging.INFO)



"""
Bumps plugin architecture.

With sophisticated models, developers need to be able to provide tools
such as model builders and data viewers.

Some of these will be tools for the GUI, such as views.  Others will be
tools to display results.

This file defines the interface that can be defined by your own application
so that it interacts with models of your type.  Define your own model
package with a module plugin.py.

Create a main program which looks like::


    if __name__ == "__main__":
        import multiprocessing
        multiprocessing.freeze_support()

        import bumps.cli
        import mypackage.plugin
        bumps.cli.install_plugin(mypackage.plugin)
        bumps.cli.main()

You should be able to use this as a driver program for your application.

Note: the plugin architecture is likely to change radically as more models
are added to the system, particularly so that we can accommodate simultaneous
fitting of data taken using different experimental techniques.  For now, only
only one plugin at a time is supported.
"""

__all__ = [
    "new_model",
    "load_model",
    "calc_errors",
    "show_errors",
    "data_view",
    "model_view",
    "migrate_serialized",
]

# TODO: refl1d wants to do the following after cli.getopts()
#
#    from refl1d.probe import Probe
#    Probe.view = opts.plot
#
# It also wants to modify the opts so that more plotters are available,
# such as Fresnel.


def new_model():
    """
    Return a new empty model or None.

    Called in response to >File >New from the GUI.  Creates a new empty
    model.  Also triggered if GUI is started without a model.
    """
    return None


def load_model(filename):
    """
    Return a model stored within a file.

    This routine is for specialized model descriptions not defined by script.

    If the filename does not contain a model of the appropriate type (e.g.,
    because the extension is incorrect), then return None.

    No need to load pickles or script models.  These will be attempted if
    load_model returns None.
    """
    return None


def calc_errors(problem, sample):
    """
    Gather data needed to display uncertainty in the model and the data.

    Returns an object to be passed later to :func:`show_errors`.
    """
    return None


def show_errors(errs, fig=None):
    """
    Display the model with uncertainty on the current figure.

    *errs* is the data returned from calc_errs.
    """
    pass


def data_view():
    """
    Panel factory for the data tab in the GUI.

    If your model has an adequate show() function this should not be
    necessary.
    """
    from .gui.data_view import DataView

    return DataView


def model_view():
    """
    Panel factory for the model tab in the GUI.

    Return None if not present.
    """
    return None


def migrate_serialized(model_dict):
    """
    Migrate serialized model to the current version.

    This function is called when loading a model from a file.  It is
    used to update the model to the current version of the plugin.
    """
    return model_dict



#!/usr/bin/python
# Copyright (C) 2006-2011, University of Maryland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author: James Krycka

"""
This module implements InputListPanel, InputListDialog, and InputListValidator
classes to provide general purpose mechanisms for obtaining and validating user
input from a structured list of input fields.
"""

import wx
from wx.lib.scrolledpanel import ScrolledPanel

from .utilities import phoenix

Validator = wx.Validator if phoenix else wx.PyValidator

WINDOW_BKGD_COLOUR = "#ECE9D8"
PALE_YELLOW = "#FFFFB0"

DATA_ENTRY_ERRMSG = """\
Please correct any highlighted field in error.
Yellow means an input value is required.
Pink indicates a syntax error."""


class ItemListValidator(Validator):
    """
    This class implements a custom input field validator.  Each instance of
    this class services one data entry field (typically implemented as
    wx.TextCtrl or a wx.ComboBox widget).  Parameters are:

    - datatype of the field (used when validating user input) as follows:
      o int       => signed or unsigned integer value
      o float     => floating point value
      o str       => string of characters
      o 'str_alpha' => string of alphabetic characters {A-Z, a-z}
      o 'str_alnum' => string of alphanumeric characters {A-Z, a-z, 0-9}
      o 'str_id'    => string identifier consisting of {A-Z, a-z, 0-9, _, -}
      o '' or any unknown datatype is treated the same as 'str'

    - flag to indicate whether user input is required (True) or optional (False)
    """

    def __init__(self, datatype="str", required=False):
        Validator.__init__(self)
        self.datatype = datatype
        self.required = required

    def Clone(self):
        # Every validator must implement the Clone() method that returns a
        # instance of the class as follows:
        return ItemListValidator(self.datatype, self.required)

    def Validate(self, win):
        """
        Verify user input according to the expected datatype.  Leading and
        trailing whitespace is always stripped before evaluation.  Floating and
        integer values are returned as normalized float or int objects; thus
        conversion can generate an error.  On error, the field is highlighted
        and the cursor is placed there.  Note that all string datatypes are
        returned stripped of leading and trailing whitespace.
        """

        text_ctrl = self.GetWindow()
        text = text_ctrl.GetValue().strip()

        try:
            if callable(self.datatype):
                self.value = self.value_alt = self.datatype(text)
            elif self.datatype == int:
                if len(text) == 0:
                    self.value = 0
                    self.value_alt = None
                else:
                    float_value = float(text)
                    if float_value != int(float_value):
                        raise ValueError("input must be an integer")
                    self.value = self.value_alt = int(float_value)
            elif self.datatype == float:
                if len(text) == 0:
                    self.value = 0.0
                    self.value_alt = None
                else:
                    self.value = self.value_alt = float(text)
            elif self.datatype == "str_alpha":
                if len(text) == 0:
                    self.value = ""
                    self.value_alt = None
                    if self.required:
                        raise RuntimeError("input required")
                else:
                    if text.isalpha():
                        self.value = self.value_alt = str(text)
                    else:
                        raise ValueError("input must be alphabetic")
            elif self.datatype == "str_alnum":
                if len(text) == 0:
                    self.value = ""
                    self.value_alt = None
                else:
                    if text.isalnum():
                        self.value = self.value_alt = str(text)
                    else:
                        raise ValueError("input must be alphanumeric")
            elif self.datatype == "str_id":
                if len(text) == 0:
                    self.value = ""
                    self.value_alt = None
                else:
                    temp = text.replace("_", "a").replace("-", "a")
                    if temp.isalnum():
                        self.value = self.value_alt = str(text)
                    else:
                        raise ValueError("input must be alphanumeric, _, or -")
            else:  # For self.datatype of "str", "", or any unrecognized type.
                if len(text) == 0:
                    self.value = ""
                    self.value_alt = None
                else:
                    self.value = self.value_alt = str(text)

            if len(text) == 0 and self.required:
                raise RuntimeError("input required")

        except RuntimeError:
            from traceback import print_exc

            print_exc()
            text_ctrl.SetBackgroundColour(PALE_YELLOW)
            text_ctrl.SetFocus()
            text_ctrl.Refresh()
            return False

        except Exception:
            from traceback import print_exc

            print_exc()
            text_ctrl.SetBackgroundColour("PINK")
            text_ctrl.SetFocus()
            text_ctrl.Refresh()
            return False

        else:
            text_ctrl.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW))
            text_ctrl.Refresh()
            self.TransferFromWindow()
            return True

    def TransferToWindow(self):
        # The parent of this class is responsible for setting the default value
        # for the field (e.g., by calling wx.TextCtrl() or wx.ComboBox() or
        # instance.SetValue(), etc.).
        return True  # Default is False for failure

    def TransferFromWindow(self):
        # Data has already been transferred from the window and validated
        # in Validate(), so there is nothing useful to do here.
        return True  # Default is False for failure

    def GetValidatedInput(self):
        # Return the validated value or zero or blank for a null input.
        return self.value

    def GetValidatedInputAlt(self):
        # Return the validated value or None for a null input.
        return self.value_alt


# ==============================================================================


class InputListPanel(ScrolledPanel):
    """
    This class implements a general purpose mechanism for obtaining and
    validating user input from several fields in a window with scroll bars.
    (See InputListDialog that uses a dialog box instead of a scrolled window.)

    It creates a scrolled window in which to display one or more input fields
    each preceded by a label.  The input fields can be a combination of simple
    data entry boxes or drop down combo boxes.  Automatic validation of user
    input is performed.  The caller can use the GetResults() method to obtain
    the final results from all fields in the form of a list of values.

    The scrolled window object is created as a child of the parent panel passed
    in.  Normally the caller of this class puts this returned object in a sizer
    attached to the parent panel to allow it to expand or contract based on the
    size constraints imposed by its parent.

    The layout is:

    +-------------------------------------+-+
    |                                     |v|
    |  Label-1:   [<drop down list>  |V]  |e|
    |                                     |r|   Note that drop down lists and
    |  Label-2:   [<data entry field-2>]  |t|   simple data entry fields can
    |  ...                                |||   be specified in any order.
    |  Label-n:   [<data entry field-n>]  |||
    |                                     |v|
    +-------------------------------------+-+   Note that scroll bars are
    |      horizontal scroll bar -->      | |   visible only when needed.
    +-------------------------------------+-+

    The itemlist parameter controls the display.  It is a list of input field
    description lists where each description list contains 5 or 6 elements and
    the 6th element is optional.  The items in the description list are:

    [0] Label string prefix for the input field
    [1] Default value
    [2] Datatype for validation (see ItemListValidator docstring for details)
    [3] Flags parameter in the form of a string of characters as follows:
        R - input is required; otherwise input is optional and can be blank
        E - field is editable by the user; otherwise it is non-editable and box
            is grayed-out; a non-editable field has its default value returned
        C - field is a combobox; otherwise it is a simple data entry box
        L - field is preceded by a divider line; 'L' takes precedent over 'H'
        H - field is preceded by a header given in the 6th element of the list;
            the following header sub-options are valid only if 'H' is specified:
            0 - header text size is same as label text size (default)
            1 - header text size is label text size + 1 point (large)
            2 - header text size is label text size + 2 points (x-large)
            3 - header text size is label text size + 3 points (2x-large)
            B - header text is bolded
            U - header text is underlined
        Options can be combined in the flags string such as 'REHB2' which means
        field is required, editable, and preceeded by a bold, extra-large header
    [4] List of values for a combo box or None for a simple data entry field
    [5] Header string to be displayed above the label string of the input field;
        if 'H' is not specified, this list element can be omitted or can be None

    The align parameter determines whether input fields are aligned across when
    the input fields are grouped into sections.  If True, the widest text label
    determines the space allocated for all labels; if False, the text label
    width is determined separately for each section.

    The fontsize parameter allows the caller to specify a font point size to
    override the default point size.

    See the AppTestFrame class for a comprehensive example.
    """

    def __init__(
        self,
        parent,
        id=wx.ID_ANY,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.TAB_TRAVERSAL,
        name="",
        itemlist=[],
        align=False,
        fontsize=None,
    ):
        ScrolledPanel.__init__(self, parent, id, pos, size, style, name)

        # self.SetBackgroundColour(WINDOW_BKGD_COLOUR)
        self.align = align
        self.itemlist = itemlist
        self.item_cnt = len(self.itemlist)
        if self.item_cnt == 0:
            return

        # Set the default font for this and all child windows (widgets) if the
        # caller specifies a size; otherwise let it default from the parent.
        if fontsize is not None:
            font = self.GetFont()
            font.SetPointSize(fontsize)
            self.SetFont(font)
        # print "Input List Panel font ptsize =", self.GetFont().GetPointSize()

        # Specify the widget layout using sizers.
        main_sizer = wx.BoxSizer(wx.VERTICAL)

        # Create the text controls for labels and associated input fields
        # and any optional headers.
        self.add_items_to_panel()

        # Divide the input items into sections prefaced by header text (except
        # that the first section is not required to have a header).  A section
        # list is created that contains the index of the item that starts a new
        # section plus a final entry that is one beyond the last item.
        sect = [0]  # declare item 0 to be start of a new section
        for i in range(self.item_cnt):
            if i > 0 and self.headers[i] is not None:
                sect.append(i)
        sect.append(self.item_cnt)

        # Place the items for each section in its own flex grid sizer.
        for i in range(len(sect) - 1):
            j = sect[i]
            k = sect[i + 1] - 1
            fg_sizer = self.add_items_to_sizer(j, k)

            # Add the flex grid sizer to the main sizer.
            if self.headers[j] is not None:  # self.headers[0] could be None
                main_sizer.Add(self.headers[j], 0, border=10, flag=wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT)
            main_sizer.Add(fg_sizer, 0, border=10, flag=wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT)

        # Finalize the sizer and establish the dimensions of the input box.
        self.SetSizer(main_sizer)
        main_sizer.Fit(self)

        # Enable scrolling and initialize the validators (required when
        # validators are not used in the context of a dialog box).
        self.SetupScrolling(scroll_x=True, scroll_y=True)
        self.InitDialog()

    def add_items_to_panel(self):
        """
        For each input item, create a header (optional), label, and input box
        widget to instantiate it.  Put the handles for these widgets in the
        headers, labels, and inputs lists where the length of each list is the
        same as the number of input boxes.
        """

        self.headers = []
        self.labels = []
        self.inputs = []
        self.widest = 0

        for x in range(self.item_cnt):
            params = len(self.itemlist[x])
            if params == 6:
                text, default, datatype, flags, plist, header = self.itemlist[x]
            elif params == 5:
                text, default, datatype, flags, plist = self.itemlist[x]
                header = None
            if default is None:
                default = ""  # display None as a null string

            # Process the flags parameter.
            required = False
            if "R" in flags:
                required = True
            editable = False
            if "E" in flags:
                editable = True
            combo = False
            if "C" in flags:
                combo = True
            line = False
            if "L" in flags:
                line = True
            hdr = False
            if "H" in flags and header is not None:
                hdr = True
            if hdr:
                delta_pts = 0
                if "1" in flags:
                    delta_pts = 1  # large
                if "2" in flags:
                    delta_pts = 2  # X-large
                if "3" in flags:
                    delta_pts = 3  # 2X-large
                weight = wx.NORMAL
                if "B" in flags:
                    weight = wx.BOLD
                underlined = False
                if "U" in flags:
                    underlined = True

            # Optionally, create a header widget to display above the input box.
            # A dividing line is treated as a special case header.
            if line:
                lin = wx.StaticLine(self, wx.ID_ANY, style=wx.LI_HORIZONTAL)
                self.headers.append(lin)
            elif hdr:
                hdr = wx.StaticText(self, wx.ID_ANY, label=header, style=wx.ALIGN_CENTER)
                font = hdr.GetFont()
                ptsize = font.GetPointSize() + delta_pts
                font.SetPointSize(ptsize)
                font.SetWeight(weight)
                font.SetUnderlined(underlined)
                hdr.SetFont(font)
                hdr.SetForegroundColour("BLUE")
                self.headers.append(hdr)
            else:
                self.headers.append(None)

            # Create the text label widget.
            self.labels.append(wx.StaticText(self, wx.ID_ANY, label=text, style=wx.ALIGN_LEFT))
            w, h = self.labels[x].GetSize()
            if w > self.widest:
                self.widest = w

            # Create the input box widget (combo box or simple data entry box)
            if combo:  # it is a drop down combo box list
                self.inputs.append(
                    wx.ComboBox(
                        self,
                        wx.ID_ANY,
                        value=str(default),
                        validator=ItemListValidator(datatype, required),
                        choices=plist,
                        style=wx.CB_DROPDOWN | wx.CB_READONLY,
                    )
                )
                self.Bind(wx.EVT_COMBOBOX, self.OnComboBoxSelect, self.inputs[x])
            else:  # it is a simple data entry field
                self.inputs.append(
                    wx.TextCtrl(self, wx.ID_ANY, value=str(default), validator=ItemListValidator(datatype, required))
                )
                self.Bind(wx.EVT_TEXT, self.OnText, self.inputs[x])

            # Verfiy that field is editable, otherwise don't allow user to edit
            if not editable:
                self.inputs[x].Enable(False)

            # Validate the default value and highlight the field if the value is
            # in error or if input is required and the value is a null string.
            self.inputs[x].GetValidator().Validate(self.inputs[x])

        # Determine if all input boxes should be aligned across sections.
        if self.align:
            for x in range(self.item_cnt):
                self.labels[x].SetMinSize((self.widest, -1))

    def add_items_to_sizer(self, start, end):
        sizer = wx.FlexGridSizer(cols=2, hgap=5, vgap=10)
        for x in range(start, end + 1):
            sizer.Add(self.labels[x], 0, wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL)
            sizer.Add(self.inputs[x], 0, wx.EXPAND)
        sizer.AddGrowableCol(1)
        return sizer

    def update_items_in_panel(self, new_values):
        for x in range(len(self.inputs)):
            if new_values[x] is not None:
                self.inputs[x].SetValue(str(new_values[x]))

    def GetResults(self):
        """
        Returns a list of values, one for each input field.  The value for
        a field is either its initial (default) value or the last value
        entered by the user that has been successfully validated.  An input
        that fails validation is not returned by the validator from the
        window.  For a non-editable field, its initial value is returned.

        Blank input is converted to 0 for int, 0.0 for float, or a 0-length
        string for a string datatype.
        """

        ret = []
        for x in range(self.item_cnt):
            ret.append(self.inputs[x].GetValidator().GetValidatedInput())
        return ret

    def GetResultsAltFormat(self):
        """
        Returns a list of values, one for each input field.  The value for
        a field is either its initial (default) value or the last value
        entered by the user that has been successfully validated.  An input
        that fails validation is not returned by the validator from the
        window.  For a non-editable field, its initial value is returned.

        Blank input is returned as a value of None.
        """

        ret = []
        for x in range(self.item_cnt):
            ret.append(self.inputs[x].GetValidator().GetValidatedInputAlt())
        return ret

    def GetResultsRawInput(self):
        """
        Returns a list of strings corresponding to each input field.  These
        are the current values from the text control widgets whether or not
        they have passed validation.  All values are returned as raw strings
        (i.e., they are not converted to floats or ints and leading and
        trailing whitespace is not stripped).
        """

        ret = []
        for x in range(self.item_cnt):
            ret.append(str(self.inputs[x].GetValue()))
        return ret

    def OnText(self, event):
        """
        This method is called each time a key stroke is entered in any text
        control box.  It should be subclassed if special processing is needed.
        The sample code below shows how to obtain the index of the box and its
        value.  Note that the box's index is 0 to n, where n is the number of
        input and combo boxes, not just the number of input boxes.

        # Get index of the input box that triggered the event.
        text_ctrl = event.GetEventObject()
        for box_idx, box_instance in enumerate(self.inputs):
            if text_ctrl is box_instance:
                break
        # Get the edited string.
        text = text_ctrl.GetValue()
        print "Field:", box_idx, text
        """

        # Run the validator bound to the text control box that has been edited.
        # If the validation fails, the validator will highlight the input field
        # to alert the user of the error.
        text_ctrl = event.GetEventObject()
        text_ctrl.GetValidator().Validate(text_ctrl)
        event.Skip()

    def OnComboBoxSelect(self, event):
        """
        This method is called each time a selection is made in any combo box.
        It should be subclassed if the caller wants to perform some action in
        response to a selection event.  The sample code below shows how to
        obtain the index of the box, the index of the item selected, and the
        value.  Note that the box's index is 0 to n, where n is the number of
        combo and input boxes, not just the number of combo boxes.

        # Get index of selected item in combo box dropdown list.
        item_idx = event.GetSelection()
        # Get index of combo box that triggered the event.
        current_box = event.GetEventObject()
        for box_idx, box_instance in enumerate(self.inputs):
            if current_box is box_instance:
                break
        print "Combo:", box_idx, item_idx, self.itemlist[box_idx][3][item_idx]
        """

        # Run the validator bound to the combo box that has a selection event.
        # This should not fail unless the combo options were setup incorrectly.
        # If the validation fails, the validator will highlight the input field
        # to alert the user of the error.
        combo_box = event.GetEventObject()
        combo_box.GetValidator().Validate(combo_box)
        event.Skip()


# ==============================================================================


class InputListDialog(wx.Dialog):
    """
    This class implements a general purpose mechanism for obtaining and
    validating user input from several fields in a pop-up dialog box.
    (See InputListPanel that uses a scrolled window instead of a dialog box.)

    It creates a pop-up dialog box in which to display one or more input fields
    each preceded by a label.  The input fields can be a combination of simple
    data entry boxes or drop down combo boxes.  Automatic validation of user
    input is performed.  OK and Cancel buttons are provided at the bottom of
    the dialog box for the user to signal completion of data entry whereupon
    the caller can use the GetResults() method to obtain the final results from
    all fields in the form of a list of values.  As with any dialog box, when
    the user presses OK or Cancel the dialog disappears from the screen, but
    the caller of this class is responsible for destroying the dialog box.

    The dialog box is automatically sized to fit the fields and buttons with
    reasonable spacing between the widgets.  The layout is:

    +-------------------------------------+
    |  Title                          [X] |
    +-------------------------------------+
    |                                     |
    |  Label-1:   [<drop down list>  |V]  |
    |                                     |     Note that drop down lists and
    |  Label-2:   [<data entry field-2>]  |     simple data entry fields can
    |  ...                                |     be specified in any order.
    |  Label-n:   [<data entry field-n>]  |
    |                                     |
    |       [  OK  ]      [Cancel]        |
    |                                     |
    +-------------------------------------+

    The itemlist parameter controls the display.  It is a list of input field
    description lists where each description list contains 5 or 6 elements and
    the 6th element is optional.  The items in the description list are:

    [0] Label string prefix for the input field
    [1] Default value
    [2] Datatype for validation (see ItemListValidator docstring for details)
    [3] Flags parameter in the form of a string of characters as follows:
        R - input is required; otherwise input is optional and can be blank
        E - field is editable by the user; otherwise it is non-editable and box
            is grayed-out; a non-editable field has its default value returned
        C - field is a combobox; otherwise it is a simple data entry box
        L - field is preceded by a divider line; 'L' takes precedent over 'H'
        H - field is preceded by a header given in the 6th element of the list;
            the following header sub-options are valid only if 'H' is specified:
            0 - header text size is same as label text size (default)
            1 - header text size is label text size + 1 point (large)
            2 - header text size is label text size + 2 points (x-large)
            3 - header text size is label text size + 3 points (2x-large)
            B - header text is bolded
            U - header text is underlined
        Options can be combined in the flags string such as 'REHB2' which means
        field is required, editable, and preceeded by a bold, extra-large header
    [4] List of values for a combo box or None for a simple data entry field
    [5] Header string to be displayed above the label string of the input field;
        if 'H' is not specified, this list element can be omitted or can be None

    The align parameter determines whether input fields are aligned across when
    the input fields are grouped into sections.  If True, the widest text label
    determines the space allocated for all labels; if False, the text label
    width is determined separately for each section.

    The fontsize parameter allows the caller to specify a font point size to
    override the default point size.

    See the AppTestFrame class for a comprehensive example.
    """

    def __init__(
        self,
        parent=None,
        id=wx.ID_ANY,
        title="Enter Data",
        pos=wx.DefaultPosition,
        size=(300, -1),  # x is min_width; y will be calculated
        style=wx.DEFAULT_DIALOG_STYLE,
        name="",
        itemlist=[],
        align=False,
        fontsize=None,
    ):
        wx.Dialog.__init__(self, parent, id, title, pos, size, style, name)

        self.align = align
        self.itemlist = itemlist
        self.item_cnt = len(self.itemlist)
        if self.item_cnt == 0:
            return

        # Set the font for this window and all child windows (widgets) from the
        # parent window, or from the system defaults if no parent is given.
        # A dialog box does not inherit font info from its parent, so we will
        # explicitly get it from the parent and apply it to the dialog box.
        if parent is not None:
            font = parent.GetFont()
            self.SetFont(font)

        # If the caller specifies a font size, override the default value.
        if fontsize is not None:
            font = self.GetFont()
            font.SetPointSize(fontsize)
            self.SetFont(font)
        # print "Input Dialog box font ptsize =", self.GetFont().GetPointSize()

        # Create the button controls (OK and Cancel) and bind their events.
        ok_button = wx.Button(self, wx.ID_OK, "OK")
        ok_button.SetDefault()
        cancel_button = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.Bind(wx.EVT_BUTTON, self.OnOk, ok_button)

        # Specify the widget layout using sizers.
        main_sizer = wx.BoxSizer(wx.VERTICAL)

        # Create the text controls for labels and associated input fields
        # and any optional headers.
        self.add_items_to_dialog_box()

        # Divide the input items into sections prefaced by header text (except
        # that the first section is not required to have a header).  A section
        # list is created that contains the index of the item that starts a new
        # section plus a final entry that is one beyond the last item.
        sect = [0]  # declare item 0 to be start of a new section
        for i in range(self.item_cnt):
            if i > 0 and self.headers[i] is not None:
                sect.append(i)
        sect.append(self.item_cnt)
        # print "Section index list:", sect

        # Place the items for each section in its own flex grid sizer.
        for i in range(len(sect) - 1):
            j = sect[i]
            k = sect[i + 1] - 1
            # print "Items per section:", j, "to", k
            fg_sizer = self.add_items_to_sizer(j, k)

            # Add the flex grid sizer to the main sizer.
            if self.headers[j] is not None:  # self.headers[0] could be None
                main_sizer.Add(self.headers[j], 0, border=10, flag=wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT)
            main_sizer.Add(fg_sizer, 0, border=10, flag=wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT)

        # Create the button sizer that will put the buttons in a row, right
        # justified, and with a fixed amount of space between them.  This
        # emulates the Windows convention for placing a set of buttons at the
        # bottom right of the window.
        button_sizer = wx.BoxSizer(wx.HORIZONTAL)
        button_sizer.Add((10, 20), 1)  # stretchable whitespace
        button_sizer.Add(ok_button, 0)
        button_sizer.Add((10, 20), 0)  # non-stretchable whitespace
        button_sizer.Add(cancel_button, 0)

        # Add a separator line before the buttons.
        separator = wx.StaticLine(self, wx.ID_ANY, style=wx.LI_HORIZONTAL)
        main_sizer.Add(separator, 0, border=10, flag=wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT)

        # Add the button sizer to the main sizer.
        main_sizer.Add(button_sizer, 0, border=10, flag=wx.EXPAND | wx.TOP | wx.BOTTOM | wx.RIGHT)

        # Finalize the sizer and establish the dimensions of the dialog box.
        # The minimum width is explicitly set because the sizer is not able to
        # take into consideration the width of the enclosing frame's title.
        self.SetSizer(main_sizer)
        main_sizer.SetMinSize((size[0], -1))
        main_sizer.Fit(self)

    def add_items_to_dialog_box(self):
        """
        For each input item, create a header (optional), label, and input box
        widget to instantiate it.  Put the handles for these widgets in the
        headers, labels, and inputs lists where the length of each list is the
        same as the number of input boxes.
        """

        self.headers = []
        self.labels = []
        self.inputs = []
        self.widest = 0
        first_error_idx = None

        for x in range(self.item_cnt):
            params = len(self.itemlist[x])
            if params == 6:
                text, default, datatype, flags, plist, header = self.itemlist[x]
            elif params == 5:
                text, default, datatype, flags, plist = self.itemlist[x]
                header = None
            if default is None:
                default = ""  # display None as a null string

            # Process the flags parameter.
            required = False
            if flags.find("R") >= 0:
                required = True
            editable = False
            if flags.find("E") >= 0:
                editable = True
            combo = False
            if flags.find("C") >= 0:
                combo = True
            line = False
            if flags.find("L") >= 0:
                line = True
            hdr = False
            if flags.find("H") >= 0 and header is not None:
                hdr = True
            if hdr:
                delta_pts = 0
                if flags.find("1") >= 0:
                    delta_pts = 1  # large
                if flags.find("2") >= 0:
                    delta_pts = 2  # X-large
                if flags.find("3") >= 0:
                    delta_pts = 3  # 2X-large
                weight = wx.NORMAL
                if flags.find("B") >= 0:
                    weight = wx.BOLD
                underlined = False
                if flags.find("U") >= 0:
                    underlined = True

            # Optionally, create a header widget to display above the input box.
            # A dividing line is treated as a special case header.
            if line:
                lin = wx.StaticLine(self, wx.ID_ANY, style=wx.LI_HORIZONTAL)
                self.headers.append(lin)
            elif hdr:
                hdr = wx.StaticText(self, wx.ID_ANY, label=header, style=wx.ALIGN_CENTER)
                font = hdr.GetFont()
                ptsize = font.GetPointSize() + delta_pts
                font.SetPointSize(ptsize)
                font.SetWeight(weight)
                font.SetUnderlined(underlined)
                hdr.SetFont(font)
                hdr.SetForegroundColour("BLUE")
                self.headers.append(hdr)
            else:
                self.headers.append(None)

            # Create the text label widget.
            self.labels.append(wx.StaticText(self, wx.ID_ANY, label=text, style=wx.ALIGN_LEFT))
            w, h = self.labels[x].GetSize()
            if w > self.widest:
                self.widest = w

            # Create the input box widget (combo box or simple data entry box)
            if combo:  # it is a drop down combo box list
                self.inputs.append(
                    wx.ComboBox(
                        self,
                        wx.ID_ANY,
                        value=str(default),
                        validator=ItemListValidator(datatype, required),
                        choices=plist,
                        style=wx.CB_DROPDOWN | wx.CB_READONLY,
                    )
                )
                self.Bind(wx.EVT_COMBOBOX, self.OnComboBoxSelect, self.inputs[x])
            else:  # it is a simple data entry field
                self.inputs.append(
                    wx.TextCtrl(self, wx.ID_ANY, value=str(default), validator=ItemListValidator(datatype, required))
                )
                self.Bind(wx.EVT_TEXT, self.OnText, self.inputs[x])

            # Verfiy that field is editable, otherwise don't allow user to edit
            if not editable:
                self.inputs[x].Enable(False)

            # Validate the default value and highlight the field if the value is
            # in error or if input is required and the value is a null string.
            # Also, save index of the first field to fail validation, if any.
            ret = self.inputs[x].GetValidator().Validate(self.inputs[x])
            if not ret and first_error_idx is None:
                first_error_idx = x

        # If any fields failed validation, set focus to the first failed one.
        if first_error_idx is not None:
            self.inputs[first_error_idx].SetFocus()

        # Determine if all input boxes should be aligned across sections.
        if self.align:
            for x in range(self.item_cnt):
                self.labels[x].SetMinSize((self.widest, -1))

    def add_items_to_sizer(self, start, end):
        sizer = wx.FlexGridSizer(cols=2, hgap=5, vgap=10)
        for x in range(start, end + 1):
            sizer.Add(self.labels[x], 0, wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL)
            sizer.Add(self.inputs[x], 0, wx.EXPAND)
        sizer.AddGrowableCol(1)
        return sizer

    def update_items_in_dialog_box(self, new_values):
        for x in range(len(self.inputs)):
            if new_values[x] is not None:
                self.inputs[x].SetValue(str(new_values[x]))

    def GetResults(self):
        """
        Returns a list of values, one for each input field.  The value for
        a field is either its initial (default) value or the last value
        entered by the user that has been successfully validated.  An input
        that fails validation is not returned by the validator from the
        window.  For a non-editable field, its initial value is returned.

        Blank input is converted to 0 for int, 0.0 for float, or a 0-length
        string for a string datatype.
        """

        ret = []
        for x in range(self.item_cnt):
            ret.append(self.inputs[x].GetValidator().GetValidatedInput())
        return ret

    def GetResultsAltFormat(self):
        """
        Returns a list of values, one for each input field.  The value for
        a field is either its initial (default) value or the last value
        entered by the user that has been successfully validated.  An input
        that fails validation is not returned by the validator from the
        window.  For a non-editable field, its initial value is returned.

        Blank input is returned as a value of None.
        """

        ret = []
        for x in range(self.item_cnt):
            ret.append(self.inputs[x].GetValidator().GetValidatedInputAlt())
        return ret

    def GetResultsRawInput(self):
        """
        Returns a list of strings corresponding to each input field.  These
        are the current values from the text control widgets whether or not
        they have passed validation.  All values are returned as raw strings
        (i.e., they are not converted to floats or ints and leading and
        trailing whitespace is not stripped).
        """

        ret = []
        for x in range(self.item_cnt):
            ret.append(str(self.inputs[x].GetValue()))
        return ret

    def OnOk(self, event):
        """
        This method gets called when the user presses the OK button.
        It is intended to be subclassed if special processing is needed.
        """

        # Explicitly validate all input values before proceeding.  Although
        # char-by-char validation would have warned the user about any invalid
        # entries, the user could have pressed the OK button without making
        # the corrections, so we'll do a full validation pass now.  The only
        # purpose is to display an explicit error if any input fails validation.
        if not self.Validate():
            wx.MessageBox(caption="Data Entry Error", message=DATA_ENTRY_ERRMSG, style=wx.ICON_ERROR | wx.OK)
            return  # keep the dialog box open

        # When the wx.ID_OK event is skipped (to allow handlers up the chain to
        # run), the Validate methods for all text control boxes will be called.
        # If all report success, the TransferFromWindow methods will be called
        # and the dialog box will close.  However, if any Validate method fails
        # this process will stop and the dialog box will remain open allowing
        # the user to either correct the problem(s) or cancel the dialog.
        event.Skip()

    def OnText(self, event):
        """
        This method is called each time a key stroke is entered in any text
        control box.  It should be subclassed if special processing is needed.
        The sample code below shows how to obtain the index of the box and its
        value.  Note that the box's index is 0 to n, where n is the number of
        input and combo boxes, not just the number of input boxes.

        # Get index of the input box that triggered the event.
        text_ctrl = event.GetEventObject()
        for box_idx, box_instance in enumerate(self.inputs):
            if text_ctrl is box_instance:
                break
        # Get the edited string.
        text = text_ctrl.GetValue()
        print "Field:", box_idx, text
        """

        # Run the validator bound to the text control box that has been edited.
        # If the validation fails, the validator will highlight the input field
        # to alert the user of the error.
        text_ctrl = event.GetEventObject()
        text_ctrl.GetValidator().Validate(text_ctrl)
        event.Skip()

    def OnComboBoxSelect(self, event):
        """
        This method is called each time a selection is made in any combo box.
        It should be subclassed if the caller wants to perform some action in
        response to a selection event.  The sample code below shows how to
        obtain the index of the box, the index of the item selected, and the
        value.  Note that the box's index is 0 to n, where n is the number of
        combo and input boxes, not just the number of combo boxes.

        # Get index of selected item in combo box dropdown list.
        item_idx = event.GetSelection()
        # Get index of combo box that triggered the event.
        current_box = event.GetEventObject()
        for box_idx, box_instance in enumerate(self.inputs):
            if current_box is box_instance:
                break
        print "Combo:", box_idx, item_idx, self.itemlist[box_idx][3][item_idx]
        """

        # Run the validator bound to the combo box that has a selection event.
        # This should not fail unless the combo options were setup incorrectly.
        # If the validation fails, the validator will highlight the input field
        # to alert the user of the error.
        combo_box = event.GetEventObject()
        combo_box.GetValidator().Validate(combo_box)
        event.Skip()


# ==============================================================================


class AppTestFrame(wx.Frame):
    """
    Interactively test both the InputListPanel and the InputListDialog classes.
    Both will display the same input fields.  Enter invalid data to verify
    char-by-char error processing.  Press the Submit and OK buttons with an
    uncorrected highlighted field in error to generate a pop-up error box.
    Resize the main window to see scroll bars disappear and reappear.
    """

    # Establish efault font and point size for test.
    FONTNAME = "Arial"
    if wx.Platform == "__WXMSW__":
        FONTSIZE = 9
    elif wx.Platform == "__WXMAC__":
        FONTSIZE = 12
    elif wx.Platform == "__WXGTK__":
        FONTSIZE = 11
    else:
        FONTSIZE = 10

    def __init__(self):
        wx.Frame.__init__(self, parent=None, id=wx.ID_ANY, title="InputListPanel Test", size=(300, 600))
        panel = wx.Panel(self, wx.ID_ANY, style=wx.RAISED_BORDER)
        panel.SetBackgroundColour("PALE GREEN")

        pt_size = panel.GetFont().GetPointSize()

        # Define fields for both InputListPanel and InputListDialog to display.
        self.fields = [
            ["Integer (int, optional):", 12345, int, "EH3", None, "Test Header (2X-large)"],
            # Test specification of integer default value as a string
            ["Integer (int, optional):", "-60", int, "E", None],
            # Default value is null, so the required field should be highlighted
            ["Integer (int, required):", "", int, "RE", None],
            ["Floating Point (float, optional):", 2.34567e-5, float, "EHB1", None, "Test Header (large, bold)"],
            ["Floating Point (float, optional):", "", float, "E", None],
            ["Floating Point (float, required):", 1.0, float, "RE", None],
            # Test unknown datatype which should be treated as 'str'
            ["String (str, optional):", "DANSE", "foo", "EHU", None, "Test Header (%dpt font, underlined)" % pt_size],
            ["String (str, reqiured):", "delete me", str, "RE", None],
            ["Non-editable field:", "Cannot be changed!", str, "", None],
            ["ComboBox String:", "Two", str, "CREL", ("One", "Two", "Three")],
            # ComboBox items must be specified as strings
            ["ComboBox Integer:", "", int, "CE", ("100", "200", "300")],
            ["String (alphabetic):", "Aa", "str_alpha", "E", None],
            ["String (alphanumeric):", "Aa1", "str_alnum", "E", None],
            ["String (A-Z, a-z, 0-9, _, -):", "A-1_a", "str_id", "E", None],
        ]

        # Create the scrolled window with input boxes.  Due to the intentionally
        # small size of the parent panel, both scroll bars should be displayed.
        self.scrolled = InputListPanel(parent=panel, itemlist=self.fields, align=True)

        # Create a button to request the popup dialog box.
        show_button = wx.Button(panel, wx.ID_ANY, "Show Pop-up Dialog Box")
        self.Bind(wx.EVT_BUTTON, self.OnShow, show_button)

        # Create a button to signal end of user edits and one to exit program.
        submit_button = wx.Button(panel, wx.ID_ANY, "Submit")
        self.Bind(wx.EVT_BUTTON, self.OnSubmit, submit_button)
        exit_button = wx.Button(panel, wx.ID_ANY, "Exit")
        self.Bind(wx.EVT_BUTTON, self.OnExit, exit_button)

        # Create a horizontal sizer for the buttons.
        button_sizer = wx.BoxSizer(wx.HORIZONTAL)
        button_sizer.Add((10, 20), 1)  # stretchable whitespace
        button_sizer.Add(submit_button, 0)
        button_sizer.Add((10, 20), 0)  # non-stretchable whitespace
        button_sizer.Add(exit_button, 0)

        # Create a vertical box sizer for the panel and layout widgets in it.
        box_sizer = wx.BoxSizer(wx.VERTICAL)
        box_sizer.Add(show_button, 0, wx.ALIGN_CENTER | wx.ALL, border=10)
        box_sizer.Add(self.scrolled, 1, wx.EXPAND | wx.ALL, border=10)
        box_sizer.Add(button_sizer, 0, wx.EXPAND | wx.BOTTOM | wx.ALL, border=10)

        # Associate the sizer with its container.
        panel.SetSizer(box_sizer)
        box_sizer.Fit(panel)

    def OnShow(self, event):
        # Display the same fields shown in the frame in a pop-up dialog box.
        pt_size = self.FONTSIZE
        self.fields[6][5] = "Test Header (%dpt font, underlined)" % pt_size
        dlg = InputListDialog(
            parent=self, title="InputListDialog Test", itemlist=self.fields, align=True, fontsize=self.FONTSIZE
        )
        if dlg.ShowModal() == wx.ID_OK:
            print("****** Dialog Box results from validated input fields:")
            print("  ", dlg.GetResults())
            print("****** Dialog Box results from validated input fields" + " (None if no input):")
            print("  ", dlg.GetResultsAltFormat())
            print("****** Dialog Box results from raw input fields:")
            print("  ", dlg.GetResultsRawInput())
        dlg.Destroy()

    def OnSubmit(self, event):
        # Explicitly validate all input parameters before proceeding.  Even
        # though char-by-char validation would have warned the user about any
        # invalid entries, the user could have pressed the Done button without
        # making the corrections, so a full validation pass is necessary.
        if not self.scrolled.Validate():
            wx.MessageBox(
                caption="Data Entry Error",
                message="Please correct the highlighted fields in error.",
                style=wx.ICON_ERROR | wx.OK,
            )
            return  # keep the dialog box open
        print("****** Scrolled Panel results from validated input fields:")
        print("  ", self.scrolled.GetResults())
        print("****** Scrolled Panel results from validated input fields" + " (None if no input):")
        print("  ", self.scrolled.GetResultsAltFormat())
        print("****** Scrolled Panel results from raw input fields:")
        print("  ", self.scrolled.GetResultsRawInput())

    def OnExit(self, event):
        # Terminate the program.
        self.Close()


# ==============================================================================

if __name__ == "__main__":
    # Interactively test both the InputListPanel and the InputListDialog classes.
    app = wx.PySimpleApp()
    frame = AppTestFrame()
    frame.Show(True)
    app.MainLoop()



import io
import csv
from typing import Literal, TypedDict, Any


def csv2dict(csvdata: str) -> dict:
    with io.StringIO(csvdata) as f:
        reader = csv.DictReader(f)
        result = dict(header=reader.fieldnames, rows=[row for row in reader], raw=csvdata)

    return result


def dict2csv(csvdict: dict) -> str:
    # No longer needed since raw csv data are added to the table data
    # dictionary in csv2dict
    with io.StringIO() as f:
        writer = csv.DictWriter(f, fieldnames=csvdict["header"])
        writer.writerows(csvdict["rows"])
        result = f.getvalue()

    return result


class CustomWebviewPlot(TypedDict):
    fig_type: Literal["plotly", "matplotlib", "table", "error"] = "plotly"
    plotdata: Any


def process_custom_plot(plot_item: CustomWebviewPlot) -> CustomWebviewPlot:
    figtype = plot_item["fig_type"]
    plot_data = plot_item["plotdata"]

    if figtype == "plotly":
        figdict = plot_data.to_dict()
    elif figtype == "matplotlib":
        import mpld3

        figdict = mpld3.fig_to_dict(plot_data)
    elif figtype == "table":
        figdict = csv2dict(plot_data)
    elif figtype == "error":
        figdict = plot_data
    else:
        figdict = dict(error=f"unrecognized plot type {figtype}")

    del plot_data  # is this necessary? Does plot_item['plot_data'] also need to be deleted?

    return CustomWebviewPlot(fig_type=figtype, plotdata=figdict)



# Copyright (C) 2006-2011, University of Maryland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Author: James Krycka

"""
This module contains utility functions and classes for the application.
"""

# ==============================================================================

import glob
import os
import sys
import time

import wx
from wx.lib import delayedresult

# CRUFT: wx 3/4
phoenix = wx.version() >= "4.0"
BitmapFromImage = wx.Bitmap if phoenix else wx.BitmapFromImage


# Text string used to compare the string width in pixels for different fonts.
# This benchmark string has 273 characters, containing 92 distinct characters
# consisting of the lowercase alpha chars in the ratio used in an English
# Scrabble(TM) set, two sets of uppercase alpha chars, two sets of digits,
# special chars with multiples of commonly used ones, and many spaces to
# approximate spacing between words in sentences and labels.
BENCHMARK_TEXT = (
    "aaaaaaaaa bb cc dddd eeeeeeeeeeee ff ggg hh iiiiiiiii j k llll mm "
    "nnnnnn oooooooo pp q rrrrrr ssss tttttt uuuu vv ww x yy z "
    "ABCD EFGH IJKL MNOP QRST UVW XYZ ABCD EFGH IJKL MNOP QRST UVW XYZ "
    "01234 56789 01234 56789 "
    "...... :::: ()()() \"\",,'' ++-- **//== {}[]<> ;|~\\_ ?!@#$%^&"
)

# The width and height in pixels of the test string using MS Windows default
# font "MS Shell Dlg 2" and a dpi of 96.
# Note: the MS Windows XP default font has the same width and height as Tahoma.
BENCHMARK_WIDTH = 1600
BENCHMARK_HEIGHT = 14

# ==============================================================================


def choose_fontsize(fontname=None):
    """
    Determines the largest font size (in points) to use for a given font such
    that the rendered width of the benchmark string is less than or equal to
    101% of the rendered width of the string on a Windows XP computer using the
    Windows default font at 96 dpi.

    The width in pixels of a rendered string is affected by the choice of font,
    the point size of the font, and the resolution of the installed font as
    measured in dots-per-inch (aka points-per-inch).
    """

    frame = wx.Frame(parent=None, id=wx.ID_ANY, title="")
    if fontname is None:
        fontname = frame.GetFont().GetFaceName()
    max_width = BENCHMARK_WIDTH + BENCHMARK_WIDTH // 100

    for fontsize in range(12, 5, -1):
        frame.SetFont(wx.Font(fontsize, wx.SWISS, wx.NORMAL, wx.NORMAL, False, fontname))
        benchmark = wx.StaticText(frame, wx.ID_ANY, label="")
        w, h = benchmark.GetTextExtent(BENCHMARK_TEXT)
        benchmark.Destroy()
        if w <= max_width:
            break

    frame.Destroy()
    return fontsize


def display_fontsize(
    fontname=None, benchmark_text=BENCHMARK_TEXT, benchmark_width=BENCHMARK_WIDTH, benchmark_height=BENCHMARK_HEIGHT
):
    """
    Displays the width in pixels of a benchmark text string for a given font
    at various point sizes when rendered on the application's output device
    (which implicitly takes into account the resolution in dpi of the font
    faces at the various point sizes).
    """

    # Create a temporary frame that we will soon destroy.
    frame = wx.Frame(parent=None, id=wx.ID_ANY, title="")

    # Set the fontname if one is given, otherwise use the system default font.
    # Get the font name even if we just set it in case the specified font is
    # not installed and the system chooses another one.
    if fontname is not None:
        frame.SetFont(wx.Font(12, wx.SWISS, wx.NORMAL, wx.NORMAL, False, fontname))
    fontname = frame.GetFont().GetFaceName()

    x, y = wx.ClientDC(frame).GetPPI()
    print("*** Benchmark text width and height in pixels = %4d %2d" % (benchmark_width, benchmark_height))
    print("*** Compare against %s font with dpi resolution of %d:" % (fontname, x))

    for fontsize in range(12, 5, -1):
        frame.SetFont(wx.Font(fontsize, wx.SWISS, wx.NORMAL, wx.NORMAL, False, fontname))
        benchmark = wx.StaticText(frame, wx.ID_ANY, label="")
        w, h = benchmark.GetTextExtent(benchmark_text)
        benchmark.Destroy()
        print("      For point size %2d, benchmark text w, h = %4d  %2d" % (fontsize, w, h))

    frame.Destroy()


def _finddata():
    patterns = ["*.png", "*.ico", "*.jpg"]
    path = resource_dir()
    files = []
    for p in patterns:
        files += glob.glob(os.path.join(path, p))
    return files


def data_files():
    """
    Return the data files associated with the package.

    The format is a list of (directory, [files...]) pairs which can be
    used directly in the py2exe setup script as::

        setup(...,
              data_files=data_files(),
              ...)
    """
    data_files = [("bumps-data", _finddata())]
    return data_files


def package_data():
    """
    Return the data files associated with the package.

    The format is a dictionary of {'fully.qualified.module', [files...]}
    used directly in the setup script as::

        setup(...,
              package_data=package_data(),
              ...)
    """
    return {"bumps.gui": _finddata()}


self_cached_path = None


def resource_dir():
    """
    Return the path to the application data.

    This is either in the environment variable BUMPS_DATA, in the
    source tree in gui/resources, or beside the executable in
    bumps-data.
    """
    # If we already found it, then we are done
    global self_cached_path
    if self_cached_path is not None:
        return self_cached_path

    # Check for data path in the environment
    key = "BUMPS_DATA"
    if key in os.environ:
        path = os.environ[key]
        if not os.path.isdir(path):
            raise RuntimeError("Path in environment %s not a directory" % key)
        self_cached_path = path
        return self_cached_path

    # Check for data path in the package
    path = os.path.abspath(os.path.join(os.path.dirname(__file__), "resources"))
    # print >>sys.stderr, "checking for resource in",path
    if os.path.isdir(path):
        self_cached_path = path
        return self_cached_path

    # Check in package root, which is where pyinstaller puts it
    root = os.path.dirname(os.path.dirname(os.path.dirname(path)))
    path = os.path.join(root, "bumps-data")
    if os.path.isdir(path):
        self_cached_path = path
        return self_cached_path

    # Check for data path next to exe/zip file.
    exepath = os.path.dirname(sys.executable)
    path = os.path.join(exepath, "bumps-data")
    # print >>sys.stderr, "checking for resource in",path
    if os.path.isdir(path):
        self_cached_path = path
        return self_cached_path

    # py2app puts the data in Contents/Resources, but the executable
    # is in Contents/MacOS.
    path = os.path.join(exepath, "..", "Resources", "bumps-data")
    # print >>sys.stderr, "checking for resource in",path
    if os.path.isdir(path):
        self_cached_path = path
        return self_cached_path

    raise RuntimeError("Could not find the Bumps data files")


def resource(filename):
    return os.path.join(resource_dir(), filename)


def get_bitmap(filename, type=wx.BITMAP_TYPE_PNG, scale_factor=16):
    """
    Returns the scaled bitmap from an image file (bmp, jpg, png) stored in
    the data directory of the package.
    """

    path = resource(filename)

    return BitmapFromImage(wx.Image(name=path, type=type).Scale(scale_factor, scale_factor))


def popup_error_message(caption, message):
    """Displays an error message in a pop-up dialog box with an OK button."""

    msg = wx.MessageDialog(None, message, caption, style=wx.ICON_ERROR | wx.OK)
    msg.ShowModal()
    msg.Destroy()


def popup_information_message(caption, message):
    """Displays an informational message in a pop-up with an OK button."""

    msg = wx.MessageDialog(None, message, caption, style=wx.ICON_INFORMATION | wx.OK)
    msg.ShowModal()
    msg.Destroy()


def popup_question(caption, message):
    """Displays a question in a pop-up dialog box with YES and NO buttons."""

    msg = wx.MessageDialog(None, message, caption, style=wx.ICON_QUESTION | wx.YES_NO)
    msg.ShowModal()
    msg.Destroy()


def popup_warning_message(caption, message):
    """Displays a warning message in a pop-up dialog box with an OK button."""

    msg = wx.MessageDialog(None, message, caption, style=wx.ICON_WARNING | wx.OK)
    msg.ShowModal()
    msg.Destroy()


# ==============================================================================


class StatusBarInfo:
    """This class writes, saves, and restores multi-field status bar text."""

    def __init__(self):
        frame = wx.FindWindowByName("AppFrame", parent=None)
        self.sb = frame.GetStatusBar()
        self.cnt = self.sb.GetFieldsCount()
        self.field = [""] * self.cnt

    def write(self, index=0, text=""):
        # Write text to the specified slot and save text locally.
        # Beware that if you use field 0, wxPython will likely overwite it.
        if index > self.cnt - 1:
            return
        self.sb.SetStatusText(text, index)
        self.field[index] = text

    def restore(self):
        # Restore saved text from fields 1 to n.
        # Note that wxPython updates field 0 with hints and other messages.
        for index in range(1, self.cnt):
            self.sb.SetStatusText(self.field[index], index)


# ==============================================================================


class ExecuteInThread:
    """
    This class executes the specified function in a separate thread and calls a
    designated callback function when the execution completes.  Control is
    immediately given back to the caller of ExecuteInThread which can execute
    in parallel in the main thread.

    Note that wx.lib.delayedresult provides a simple interface to threading
    that does not include mechanism to stop the thread.
    """

    def __init__(self, callback, function, *args, **kwargs):
        if callback is None:
            callback = self._callback
        # print "*** ExecuteInThread init:", callback, function, args, kwargs
        delayedresult.startWorker(consumer=callback, workerFn=function, wargs=args, wkwargs=kwargs)

    def _callback(self, delayedResult):
        """
        jobID = delayedResult.getJobID()
        assert jobID == self.jobID
        try:
            result = delayedResult.get()
        except Exception, e:
            popup_error_message(self, "job %s raised exception: %s"%(jobID, e)
            return
        """
        return


# ==============================================================================


class WorkInProgress(wx.Panel):
    """
    This class implements a rotating 'work in progress' gauge.
    """

    def __init__(self, parent):
        wx.Panel.__init__(self, parent, wx.ID_ANY)

        self.gauge = wx.Gauge(self, wx.ID_ANY, range=50, size=(250, 25))

        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.TimerHandler)
        # self.count = 0

    def Start(self):
        self.timer.Start(100)

    def Stop(self):
        self.timer.Stop()

    def TimerHandler(self, event):
        # self.count += 1
        # print "*** count = ", self.count
        self.gauge.Pulse()


# ==============================================================================

log_time_handle = None  # global variable for holding TimeStamp instance handle


def log_time(text=None, reset=False):
    """
    This is a convenience function for using the TimeStamp class from any
    module in the application for logging elapsed and delta time information.
    This data is prefixed by a timestamp and optionally suffixed by a comment.
    log_time maintains a single instance of TimeStamp during program execution.
    Example output from calls to log_time('...'):

    ==>    0.000s    0.000s  Starting <application name>
    ==>    0.016s    0.016s  Starting to display the splash screen
    ==>    0.015s    0.031s  Starting to build the GUI application
    ==>    0.094s    0.125s  Entering the event loop
    ==>    2.906s    3.031s  Terminating the splash screen and showing the GUI
    """

    global log_time_handle
    if log_time_handle is None:
        log_time_handle = TimeStamp()
    if reset:
        log_time_handle.reset()
    log_time_handle.log_interval(text=text)


class TimeStamp:
    """
    This class provides timestamp, delta time, and elapsed time services for
    displaying wall clock time usage by the application.
    """

    def __init__(self):
        self.reset()

    def reset(self):
        # Starts new timing interval.
        self.t0 = self.t1 = time.time()

    def gettime3(self):
        # Gets current time in timestamp, delta time, and elapsed time format.
        now = time.time()
        timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(now))
        elapsed = now - self.t0
        delta = now - self.t1
        self.t1 = now
        return timestamp, delta, elapsed

    def gettime2(self):
        # Gets current time in delta time and elapsed time format.
        now = time.time()
        elapsed = now - self.t0
        delta = now - self.t1
        self.t1 = now
        return delta, elapsed

    def log_time_info(self, text=""):
        # Prints timestamp, delta time, elapsed time, and optional comment.
        t, d, e = self.gettime3()
        print("==> %s%9.3fs%9.3fs  %s" % (t, d, e, text))

    def log_timestamp(self, text=""):
        # Prints timestamp and optional comment.
        t, d, e = self.gettime3()
        print("==> %s  %s" % (t, text))

    def log_interval(self, text=""):
        # Prints elapsed time, delta time, and optional comment.
        d, e = self.gettime2()
        print("==>%9.3fs%9.3fs  %s" % (d, e, text))


# ==============================================================================

if __name__ == "__main__":
    # Test the display_fontsize and choose_fontsize functions.
    app = wx.PySimpleApp()
    print("For Arial font:")
    display_fontsize(fontname="Arial")
    print("    Calculated font size =", choose_fontsize(fontname="Arial"))
    app.Destroy()

    print("")
    print("*** Data directory is:          ", resource_dir())

    # Test the TimeStamp class and the convenience function.
    print("")
    log_time("Using log_time() function")
    print("Sleeping for 0.54 seconds ...")
    time.sleep(0.54)
    log_time("Using log_time() function")
    print("Sleeping for 0.83 seconds ...")
    time.sleep(0.83)
    log_time("Using log_time() function")
    print("Creating an instance of TimeStamp (as the second timing class)")
    ts = TimeStamp()
    print("Sleeping for 0.66 seconds ...")
    time.sleep(0.66)
    ts.log_time_info(text="Using log_time_info() method")
    ts.log_timestamp(text="Using log_timestamp() method")
    ts.log_interval(text="Using log_interval() method")
    print("Sleeping for 0.35 seconds ...")
    time.sleep(0.35)
    ts.log_interval(text="Using log_interval() method")
    print("Sleeping for 0.42 seconds ...")
    time.sleep(0.42)
    ts.log_interval(text="Using log_interval() method")
    print("Resetting the clock ...")
    ts.reset()
    ts.log_interval(text="Using log_interval() method")
    print("Sleeping for 0.33 seconds ...")
    time.sleep(0.33)
    ts.log_interval(text="Using log_interval() method")
    print("Switch back to the first timing class")
    log_time("Using log_time() function")



# This program is public domain
# Author: Paul Kienzle
"""
Fitting parameter objects.

Parameters are a big part of the interface between the model and the fitting
engine.  By saving and retrieving values and ranges from the parameter, the
fitting engine does not need to be aware of the structure of the model.

Users can also perform calculations with parameters, tying together different
parts of the model, or different models.
"""

# __all__ = [ 'Parameter']
import operator
import sys
import builtins
from dataclasses import dataclass, field, InitVar
from functools import reduce
import warnings
from copy import copy
import uuid
from functools import wraps
from enum import Enum

from typing import Type, TypeVar, Optional, Any, Union, Dict, Callable, Tuple, List, Sequence
from .util import Literal

import numpy as np
from numpy import inf, isinf, isfinite

from . import bounds as mbounds
from . import pmath
from .util import field_desc, schema_config

BoundsType = mbounds.BoundsType

ValueType = Union["Expression", "Parameter", "Calculation", float]

# TODO: avoid evaluation of subexpressions if parameters do not change.
# This is especially important if the subexpression invokes an expensive
# calculation via a parameterized function.  This will require a restructuring
# of the parameter claas.  The park-1.3 solution is viable: given a parameter
# set, figure out which order the expressions need to be evaluated by
# building up a dependency graph.  With a little care, we can check which
# parameters have actually changed since the last calculation update, and
# restrict the dependency graph to just them.
# TODO: support full aliasing, so that floating point model attributes can
# be aliased to a parameter.  The same technique as subexpressions applies:
# when the parameter is changed, the model will be updated and will need
# to be re-evaluated.


# TODO: maybe move this to util?
def to_dict(p):
    if hasattr(p, "to_dict"):
        return p.to_dict()
    elif isinstance(p, (tuple, list)):
        return [to_dict(v) for v in p]
    elif isinstance(p, dict):
        return {k: to_dict(v) for k, v in p.items()}
    elif isinstance(p, (bool, str, float, int, type(None))):
        return p
    elif isinstance(p, np.ndarray):
        # TODO: what about inf, nan and object arrays?
        return p.tolist()
    elif False and callable(p):
        # TODO: consider including functions and arbitrary values
        import base64
        import dill

        encoding = base64.encodebytes(dill.dumps(p)).decode("ascii")
        return {"type": "dill", "value": str(p), "encoding": encoding}
        ## To recovert the function
        # if allow_unsafe_code:
        #     encoding = item['encoding']
        #     p = dill.loads(base64.decodebytes(encoding).encode('ascii'))
    else:
        # print(f"converting type {type(p)} to str")
        return str(p)


@dataclass(init=False)
class Uniform:
    """Uniform distribution with hard boundaries"""


@dataclass(init=False)
class Normal:
    """Normal distribution (Gaussian)"""

    std: float = field_desc("standard deviation (1-sigma)")
    mean: float = field_desc("center of the distribution")

    def __init__(self, std: float, mean: float):
        self.std = std
        self.mean = mean


# Leave out of schema for now.
# TODO: determine if this is used by anyone
# @dataclass(init=False)
class UniformSoftBounded:
    """Uniform distribution with error-function PDF on boundaries"""

    std: float = field_desc("width of the edge distribution")


DistributionType = Union[Uniform, Normal]  # , UniformSoftBounded]


class OperatorMixin:
    """
    The set of operations that can be performed on parameter-like objects
    Parameter, Constant, Expression.

    These include: +, -, *, /, //, **, abs, float, int

    Also, numpy math functions: sin, cos, tan, ...

    Much like abs(obj) => obj.__abs__(), np.sin(obj) => obj.sin()
    """

    # float(value) is special: it returns the current value rather than
    # becoming part of the parameter expression.
    value: float

    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)

    def __bool__(self):
        # Note: __bool__ must return true or false, so we can't handle
        # lazy constraint expressions like not a, a or b, a and b.
        raise TypeError("use (p != 0) to test against zero")

    ...  # operators and functions will be filled in later


class ValueProtocol(OperatorMixin):
    """
    Values can be combined to form expressions
    Provide a suite of operators for creating parameter expressions.
    """

    fittable: bool = False
    fixed: bool = True
    value: float

    # TODO: Do values have names? Or do the names belong to the model parameter?
    # name: str
    # TODO: are priors on the parameter or on the value?
    # bounds: Optional[BoundsType] = None
    def parameters(self) -> List["Parameter"]:
        # default implementation:
        return []


@dataclass(init=False)
class Calculation(ValueProtocol):  # the name Function is taken (though deprecated)
    """
    A Parameter with a model-specific, calculated value.
    The function used to calculate this value should be well-documented in the
    description field, e.g.
    Stack.thickness: description = "a sum of the thicknesses of all layers in the stack"

    """

    description: str
    _function: Callable[[], float]  # added by the model; not serialized

    def __init__(self, description: str = ""):
        self.description = description

    @property
    def value(self):
        return self._function()

    def __float__(self):
        return self.value

    def set_function(self, function):
        self._function = function


class SupportsPrior:
    prior: Optional[BoundsType]
    limits: Tuple[float, float]
    distribution: DistributionType
    bounds: BoundsType

    def reset_prior(self):
        self.prior = None

    def has_prior(self):
        return (
            self.prior is not None
            and not isinstance(self.prior, mbounds.Unbounded)
            and self.prior.limits != (-np.inf, np.inf)
        )

    def add_prior(
        self,
        distribution: Optional[DistributionType] = None,
        bounds: Optional[BoundsType] = None,
        limits: Optional[Tuple[float, float]] = None,
    ):
        # use self values if they are found:
        if distribution is None and self.distribution is not None:
            distribution = self.distribution
        if bounds is None and self.bounds is not None:
            bounds = self.bounds
        if limits is None:
            if self.limits is not None:
                limits = self.limits
            else:
                limits = (-inf, inf)

        if bounds is not None:
            # get the intersection of the limits here.
            limits = (np.clip(limits[0], *bounds), np.clip(limits[1], *bounds))
        if isinstance(distribution, Normal):
            if limits == (-inf, inf):
                prior = mbounds.BoundedNormal(mean=distribution.mean, std=distribution.std, limits=limits)
            else:
                prior = mbounds.Normal(mean=distribution.mean, std=distribution.std)
        elif isinstance(distribution, UniformSoftBounded):
            lo, hi = limits
            prior = mbounds.SoftBounded(lo=lo, hi=hi, std=distribution.std)
        elif isinstance(distribution, Uniform):
            lo, hi = limits
            if isinf(lo) and isinf(hi):
                prior = mbounds.Unbounded()
            elif isinf(lo):
                prior = mbounds.BoundedAbove(hi)
            elif isinf(hi):
                prior = mbounds.BoundedBelow(lo)
            else:
                prior = mbounds.Bounded(lo, hi)
        else:
            raise ValueError("no distribution found matching %s" % (str(distribution)))

        self.prior = prior


@schema_config()
@dataclass(init=False)
class Parameter(ValueProtocol, SupportsPrior):
    """
    A parameter is a container for a symbolic value.

    Parameters have a prior probability, as set by a bounds constraint:

        import numpy as np
        from scipy.stats.distributions import lognorm
        from bumps.parameter import Parameter

        p = Parameter(3)
        p.pmp(10)               # 3 +/- 10% uniform
        p.pmp(-5,10)            # 3 in [2.85, 3.30] uniform
        p.pm(2)                 # 3 +/- 2 uniform
        p.pm(-1,2)              # 3 in [2,5] uniform
        p.range(0,5)            # 3 in [0,5] uniform
        p.dev(2)                # 3 +/- 2 gaussian
        p.soft_range(2,5,2)     # 3 in [2,5] uniform with gauss wings
        p.dev(2, limits=(0,6))  # 3 +/- 2 truncated gaussian
        p.pdf(lognorm(3, 1))    # lognormal centered on 3, width 1.

    Parameters have hard limits on the possible values, dictated by the model.
    These bounds apply in addition to any other bounds.

    Parameters can be constrained to be equal to another parameter or
    parameter expression:

        a, b = Parameter(3), Parameter(4)
        p = Parameter(limits=(6, 10))
        p.equals(a+b)
        assert p.nllf() == 0.   # within the bounds
        a.value = 20
        assert np.isinf(p.nllf()) # out of bounds

    Constraints on the computed value follow from the constraints on the
    underlying parameters in addition to any hard limits on the parameter
    value given by the model.

    **Inputs**

    *value* can be a constant, a variable, an expression or a link to
    another parameter.

    *bounds* are user-supplied limits on the parameter value within the model.
    If bounds are supplied then the parameter defaults to fittable.

    *distribution* is one of Uniform, Normal or UniformSoftBounded classes

    *fixed* is True if the parameter is fixed, even if bounds are supplied.

    *name* is the label associated with the parameter in plots. The names
    need not be unique, but it will be confusing if there are duplicates.
    The name will usually correspond to the role of the parameter in the
    model. For models with sequences (e.g., layer numbers), try using a
    layer name (e.g., based on the material in the layer) rather than a layer
    number for parameters in that layer. This will make it easier for the
    user to associate the parameters displayed at the end of the the fit
    with the layer in the model. Also, when exploring the space of models,
    the parameter names will be preserved even if a new layer is introduced
    before the existing layers. That will allow the parameters from the
    previous fit to be easily used as a seed for the fit to the new model.

    *id* must be a unique identifier associated with the parameter. This
    is used to link parameters on save and reload.

    *limits* are hard limits on the parameter value within the model. Separate
    from the prior distribution on a random variable provided by the user,
    the hard limits are restrictions on the value imposed by the model.
    For example, the thickness of a layer must be zero or more.

    Any additional keyword arguments are preserved as properties of the
    parameter. For example, *tip* and *units* for decorating an input form
    in the GUI:

         p = Parameter(10, name="width", units="cm", tip="Width of sample")

    """

    # Parameters may be dependent on other parameters, and the
    # fit engine will need to access them.
    # prior: Optional[BoundsType]
    id: str = field(metadata={"format": "uuid"})
    name: Optional[str] = field(default=None, init=False)
    fixed: bool = True
    slot: Union["Variable", ValueType]
    limits: Tuple[Union[float, Literal["-inf"]], Union[float, Literal["inf"]]] = (-inf, inf)
    bounds: Optional[Tuple[Union[float, Literal["-inf"]], Union[float, Literal["inf"]]]] = None
    distribution: DistributionType = field(default_factory=Uniform)
    discrete: bool = False
    tags: List[str] = field(default_factory=list)

    _fixed: bool

    def parameters(self):
        pars = [self]
        if hasattr(self.slot, "parameters"):
            pars += self.slot.parameters()
        return pars

    def pmp(self, plus, minus=None, limits=None):
        """
        Allow the parameter to vary as value +/- percent.

        pmp(*percent*) -> [value*(1-percent/100), value*(1+percent/100)]

        pmp(*plus*, *minus*) -> [value*(1+minus/100), value*(1+plus/100)]

        In the *plus/minus* form, one of the numbers should be plus and the
        other minus, but it doesn't matter which.

        If *limits* are provided, bound the end points of the range to lie
        within the limits.

        The resulting range is converted to "nice" numbers.
        """
        bounds = mbounds.pmp(self.value, plus, minus, limits=limits)
        self.bounds = bounds
        self.fixed = False
        return self

    def pm(self, plus, minus=None, limits=None):
        """
        Allow the parameter to vary as value +/- delta.

        pm(*delta*) -> [value-delta, value+delta]

        pm(*plus*, *minus*) -> [value+minus, value+plus]

        In the *plus/minus* form, one of the numbers should be plus and the
        other minus, but it doesn't matter which.

        If *limits* are provided, bound the end points of the range to lie
        within the limits.

        The resulting range is converted to "nice" numbers.
        """
        bounds = mbounds.pm(self.value, plus, minus, limits=limits)
        self.bounds = bounds
        self.fixed = False
        return self

    def dev(self, std, mean=None, limits=None, sigma=None, mu=None):
        """
        Allow the parameter to vary according to a normal distribution, with
        deviations from the mean added to the overall cost function for the
        model.

        If *mean* is None, then it defaults to the current parameter value.

        If *limits* are provide, then use a truncated normal distribution.

        Note: *sigma* and *mu* have been replaced by *std* and *mean*, but
        are left in for backward compatibility.
        """
        if sigma is not None or mu is not None:
            # CRUFT: remove sigma and mu parameters
            warnings.warn(DeprecationWarning("use std,mean instead of mu,sigma in Parameter.dev"))
            if sigma is not None:
                std = sigma
            if mu is not None:
                mean = mu
        if mean is None:
            mean = self.value  # Note: value is an attribute of the derived class
        self.bounds = limits if limits is not None else (-inf, inf)
        self.distribution = Normal(mean=mean, std=std)
        self.fixed = False
        return self

    # def pdf(self, dist):
    #     """
    #     Allow the parameter to vary according to any continuous scipy.stats
    #     distribution.
    #     """
    #     # TODO: have to make some kind of registry for distributions?
    #     # this will not work in new system of setting priors in model_reset.
    #     self._set_bounds((-inf, inf))
    #     self.distribution = dist
    #     return self

    def range(self, low, high):
        """
        Allow the parameter to vary within the given range.
        """
        self.bounds = (low, high)
        self.distribution = Uniform()
        self.fixed = False
        return self

    def soft_range(self, low, high, std):
        """
        Allow the parameter to vary within the given range, or with Gaussian
        probability, stray from the range.
        """
        self.bounds = (low, high)
        self.distribution = UniformSoftBounded(std=std)
        self.fixed = False
        return self

    # Delegate to slots
    @property
    def value(self):
        return int(self.slot) if self.discrete else float(self.slot)

    @value.setter
    def value(self, update):
        self.slot.value = round(update) if self.discrete else update

    @property
    def fittable(self):
        return isinstance(self.slot, Variable)

    @property
    def fixed(self):
        return not self.fittable or self._fixed

    @fixed.setter
    def fixed(self, state):
        # Can't set fixed to false if the parameter is not fittable
        if self.fittable:
            self._fixed = state
        elif not state:
            raise TypeError(f"value in {self.name} is not fittable")

    ## Use the following if bounds are on the value rather than the parameter
    # @property
    # def bounds(self):
    #    return getattr(self.slot, 'bounds', None)
    # @bounds.setter
    # def bounds(self, b):
    #    if not hasattr(self.slot, 'bounds'):
    #        raise TypeError(f"{self.name} is not fittable so bounds can't be set")
    #    if self.slot.fittable:
    #        self.slot.fixed = (b is None)
    #    self.slot.bounds = b

    # Functional form of parameter value access
    def __call__(self):
        return self.value

    def __float__(self):
        return float(self.value)

    def nllf(self) -> float:
        """
        Return -log(P) for the current parameter value.
        """
        value = self.value
        if not (self.limits[0] <= value <= self.limits[1]):
            # quick short-circuit if not meeting own limits:
            return np.inf
        else:
            logp = self.prior.nllf(value)
            if hasattr(self.slot, "nllf"):
                logp += self.slot.nllf()
            return logp

    def residual(self) -> float:
        """
        Return the z score equivalent for the current parameter value.

        That is, the given the value of the parameter in the underlying
        distribution, find the equivalent value in the standard normal.
        For a gaussian, this is the z score, in which you subtract the
        mean and divide by the standard deviation to get the number of
        sigmas away from the mean.  For other distributions, you need to
        compute the cdf of value in the parameter distribution and invert
        it using the ppf from the standard normal distribution.
        """
        return 0.0 if self.prior is None else self.prior.residual(self.value)

    def valid(self):
        """
        Return true if the parameter is within the valid range.
        """
        return not isinf(self.nllf())

    def format(self):
        """
        Format the parameter, value and range as a string.
        """
        return "%s=%g in %s" % (self, self.value, self.prior)

    def __str__(self):
        name = self.name if self.name is not None else "?"
        return name

    def __repr__(self):
        return "Parameter(%s)" % self

    # TODO: deprecate
    @classmethod
    def default(cls: type, value: Union[float, Tuple[float, float], ValueType], **kw) -> "Parameter":
        """
        Create a new parameter with the *value* and *kw* attributes. If value
        is already a parameter or expression, set it to that value.
        """
        # Need to constrain the parameter to fit within fixed limits and
        # to receive a name if a name has not already been provided.
        if isinstance(value, ValueProtocol):
            return value
        else:
            return cls(value, **kw)

    def set(self, value):
        """
        Set a new value for the parameter, ignoring the bounds.
        """
        self.slot.value = value

    def clip_set(self, value):
        """
        Set a new value for the parameter, clipping it to the bounds.
        """
        low, high = self.prior.limits
        self.slot.value = builtins.min(builtins.max(value, low), high)

    def __init__(
        self,
        value: Optional[Union[float, Tuple[float, float]]] = None,
        slot: Optional[Union["Variable", ValueType]] = None,
        # bounds: Optional[Union[BoundsType, Tuple[float, float]]]=None,
        fixed: Optional[bool] = None,
        name: Optional[str] = None,
        id: Optional[str] = None,
        limits: Optional[Tuple[Union[float, Literal[None, "-inf"]], Union[float, Literal[None, "inf"]]]] = None,
        bounds: Optional[Tuple[Union[float, Literal["-inf"]], Union[float, Literal["inf"]]]] = None,
        distribution: DistributionType = Uniform(),
        discrete: bool = False,
        tags: Optional[List[str]] = None,
        **kw,
    ):
        # Check if we are started with value=range or bounds=range; if we
        # are given bounds, then assume this is a fitted parameter, otherwise
        # the parameter defaults to fixed; if value is not set, use the
        # midpoint of the range.
        if bounds is None:
            try:
                # Note: throws TypeError if not a sequence (which we want to
                # fall through to the remainder of the function), or ValueError
                # if the sequence is the wrong length (which we want to fail).
                lo, hi = value
                warnings.warn(DeprecationWarning("parameters can no longer be initialized with a fit range"))
                bounds = lo, hi
                value = None
            except TypeError:
                pass
        if fixed is None:
            fixed = bounds is None
        if slot is None:
            if value is None:
                value = float(bounds[0]) if bounds is not None else 0  # ? what else to do here?
            if isinstance(value, (float, int)):
                value = round(value) if discrete else value
                slot = Variable(value)
            elif isinstance(value, ValueProtocol):
                slot = value
            else:
                raise TypeError("value %s: %s cannot be converted to Variable" % (str(name), str(value)))
        assert isinstance(slot, (float, Variable, Expression, Parameter, Constant, Calculation))

        self.slot = slot
        self.name = name
        self.id = id if id is not None else str(uuid.uuid4())
        self.tags = tags if tags is not None else []
        if limits is None:
            limits = (-np.inf, np.inf)
        self.limits = (
            (-np.inf if limits[0] is None else float(limits[0])),
            (np.inf if limits[1] is None else float(limits[1])),
        )
        if bounds is not None:
            bounds = (
                (-np.inf if bounds[0] is None else float(bounds[0])),
                (np.inf if bounds[1] is None else float(bounds[1])),
            )
        self.bounds = bounds
        self.distribution = distribution
        # Note: fixed is True unless fixed=False or bounds=bounds were given
        # as function arguments. Note that _set_bounds() will always set the
        # fixed to False, so we need to reset it after calling _set_bounds().
        self.fixed = fixed
        self.discrete = discrete

        # Store whatever values the user needs to associate with the parameter.
        # For example, models can set units and tool tips so the user interface
        # has something to work with.
        for k, v in kw.items():
            setattr(self, k, v)
        self.prior = None  # to be filled by model_reset

    def randomize(self, rng=None):
        """
        Set a random value for the parameter.
        """
        self.value = self.prior.random(rng if rng is not None else mbounds.RNG)

    def feasible(self):
        """
        Value is within the limits defined by the model
        """
        return self.prior.limits[0] <= self.value <= self.prior.limits[1]

    def equals(self, expression: ValueType):
        """
        Set a parameter equal to another parameter or expression.

        If *expression=None* then free the parameter by giving it is own
        slot with value equal to the present value of the expression, and
        its bounds.
        """
        if isinstance(self.slot, Calculation):
            raise TypeError("parameter is calculated by the model and cannot be changed")
        elif expression is self:
            # don't make a circular reference to self.
            warnings.warn(f"{self} tried to make circular reference to self...")
            pass
        else:
            self.slot = expression

    def unlink(self):
        if isinstance(self.slot, Calculation):
            raise TypeError("parameter is calculated by the model and cannot be changed")
        # Replace the slot with a new variable initialized to the only variable value
        self.slot = Variable(self.value)

    def add_tag(self, tag: str):
        if not tag in self.tags:
            self.tags.append(tag)

    def remove_tag(self, tag: Optional[str] = None):
        if tag is None:
            self.tags = []
        else:
            self.tags = [t for t in self.tags if not t == tag]

    def __copy__(self):
        """copy will only be called when a new instance is desired, with a different id"""
        obj = type(self).__new__(self.__class__)
        obj.__dict__.update(self.__dict__)
        obj.id = str(uuid.uuid4())
        return obj


def tag_all(parameter_tree, tag, remove=False):
    if isinstance(parameter_tree, dict):
        tag_all([item for item in parameter_tree.values()], tag, remove=remove)
    elif hasattr(parameter_tree, "add_tag"):
        if remove:
            parameter_tree.remove_tag(tag)
        else:
            parameter_tree.add_tag(tag)
    elif hasattr(parameter_tree, "parameters"):
        tag_all(parameter_tree.parameters(), tag, remove=remove)
    elif hasattr(parameter_tree, "__iter__"):
        for item in parameter_tree:
            tag_all(item, tag, remove=remove)
    else:
        warnings.warn(f"parameter tree should have only list, object and Parameter items: {parameter_tree}")


def untag_all(parameter_tree, tag: Optional[str] = None):
    tag_all(parameter_tree, tag, remove=True)


@dataclass
class Variable(ValueProtocol):
    """
    Saved state for a random variable in the model.
    """

    value: float

    def parameters(self):
        return []


@schema_config()
@dataclass(init=True, frozen=True, eq=False)
class Constant(ValueProtocol):  # type: ignore
    """
    Saved state for an unmodifiable value.

    A constant is like a fixed parameter. You can't change it's value, set
    it equal to another parameter, or assign a prior distribution.
    """

    value: float
    name: Optional[str] = None
    id: str = field(metadata={"format": "uuid"}, default_factory=lambda: str(uuid.uuid4()))

    fittable = False  # class property fixed across all objects
    fixed = True  # class property fixed across all objects

    def parameters(self):
        return [self]

    def __str__(self):
        return self.name


# ==== Arithmetic operators ===
class Operators(str, Enum):
    """Operators that can be used to construct Expressions"""

    # operators including abs() are defined in _build_operator_mixin()
    # functions are defined in numpy or in UserFunction (for min/max)

    # unary operator
    neg = "neg"
    pos = "pos"
    # binary operator
    add = "add"
    sub = "sub"
    mul = "mul"
    truediv = "truediv"
    floordiv = "floordiv"
    pow = "pow"
    # unary functional
    # float = "float"  => float makes values concrete
    # int = "int"  => values must be float; use floor, trunc, ceil, round
    abs = "abs"

    # unary functions
    exp = "exp"
    expm1 = "expm1"
    log = "log"
    log10 = "log10"
    log1p = "log1p"
    sqrt = "sqrt"
    degrees = "degrees"
    radians = "radians"
    sin = "sin"
    cos = "cos"
    tan = "tan"
    arcsin = "arcsin"
    arccos = "arccos"
    arctan = "arctan"
    sinh = "sinh"
    cosh = "cosh"
    tanh = "tanh"
    arcsinh = "arcsinh"
    arccosh = "arccosh"
    arctanh = "arctanh"
    ceil = "ceil"
    floor = "floor"
    trunc = "trunc"
    rint = "rint"
    round = "round"  # round(a) => rint(a)
    # binary functions
    arctan2 = "arctan2"

    # n-ary
    min = "min"  # from builtins
    max = "max"  # from builtins
    # TODO: support sum(seq) and prod(seq) for tuple and list


# Precedence for the python operators as given in manual. Numbers start
# from one at the bottom of the table. The value itself is "highest" precedence
# with a value of zero.
# https://docs.python.org/3/reference/expressions.html#operator-precedence
VALUE_PRECEDENCE = 0
CALL_PRECEDENCE = 2
OPERATOR_PRECEDENCE = {
    "pow": 4,
    "pos": 5,
    "neg": 5,
    "mul": 6,
    "truediv": 6,
    "floordiv": 6,
    "add": 7,
    "sub": 7,
    "gt": 12,
    "lt": 12,
    "ge": 12,
    "le": 12,
    "eq": 12,
    "ne": 12,
}
OPERATOR_STRING = {
    "pow": "**",
    "pos": "+",
    "neg": "-",
    "mul": "*",
    "truediv": "/",
    "floordiv": "//",
    "add": "+",
    "sub": "-",
    "gt": ">",
    "lt": "<",
    "ge": ">=",
    "le": "<=",
    "eq": "==",
    "ne": "!=",
}


def _lookup_operator(op_name):
    if not hasattr(Operators, op_name) and op_name not in UserFunctionRegistry:
        raise ValueError(f"function {op_name} is not available")
    fn = None
    # Check plugins first so we can override lookups in operator and numpy.
    # This is needed for min/max.
    if fn is None:  # plugin functions from UserFunctionRegistry
        fn = UserFunctionRegistry.get(op_name, None)
    if fn is None:
        fn = getattr(operator, op_name, None)  # operators from operators
    if fn is None:  # math functions from numpy
        fn = getattr(np, op_name, None)
    if fn is None:
        raise RuntimeError(f"should not be here: {op_name} not found")
    return fn


def _precedence(obj: Any) -> int:
    """
    Return operator precedence according to the python parsing hierarchy.

    Lower values are higher precedence. Values start at 0 for constants and
    variables, and go up from there. Not all operators are covered.
    """
    if isinstance(obj, Expression):
        return OPERATOR_PRECEDENCE.get(obj.op.name, CALL_PRECEDENCE)
    return VALUE_PRECEDENCE


@dataclass(init=False)
class Expression(ValueProtocol):
    """
    Parameter expression
    """

    fittable = False
    fixed = True

    op: Union[Operators, "UserFunction"]  # Enumerated str type {function_name: display_name}
    args: Sequence[ValueType]
    _fn: Callable[..., float]  # _fn(float, float, ...) -> float

    def __init__(self, op: Union[str, Operators, "UserFunction"], args):
        op = op if (isinstance(op, Operators) or isinstance(op, UserFunction)) else getattr(Operators, op)
        object.__setattr__(self, "op", op)
        object.__setattr__(self, "_fn", _lookup_operator(op.name))
        object.__setattr__(self, "args", args)

    def parameters(self):
        # Walk expression tree combining parameters from each subexpression
        return sum((v.parameters() for v in self.args if hasattr(v, "parameters")), [])

    @property
    def value(self):
        return self._fn(*(float(arg) for arg in self.args))

    @property
    def name(self):
        return str(self)

    def __str__(self):
        prec = _precedence(self)
        vals = [str(v) for v in self.args]
        if self.op.name in ("pos", "neg"):
            # +- a with parens as needed
            a = f"({vals[0]})" if prec < _precedence(self.args[0]) else vals[0]
            return f"{OPERATOR_STRING[self.op.name]}{a}"
        elif self.op.name in ("add", "sub", "mul", "div", "truediv", "pow"):
            # a +-*/** b with parens as needed
            a = f"({vals[0]})" if prec < _precedence(self.args[0]) else vals[0]
            b = f"({vals[1]})" if prec < _precedence(self.args[1]) else vals[1]
            return f"{a} {OPERATOR_STRING[self.op.name]} {b}"
        else:
            # f(a, b, ...) with no parens needed
            return f"{self.op.name}({', '.join(v for v in vals)})"


def _make_unary_op(op_name: str):
    op = getattr(Operators, op_name)
    # Note: self is Parameter or Expression
    fn = lambda self: Expression(op, (self,))
    setattr(OperatorMixin, f"__{op_name}__", fn)


def _make_binary_op(op_name: str):
    op = getattr(Operators, op_name)

    def fn(self, other):
        return Expression(op, (self, other))

    setattr(OperatorMixin, f"__{op_name}__", fn)

    def rfn(self, other):
        return Expression(op, (other, self))

    setattr(OperatorMixin, f"__r{op_name}__", rfn)


def _make_math_fn(fn_name: str):
    op = getattr(Operators, fn_name)

    def fn(*args):  # first of args is self
        if any([isinstance(arg, ValueProtocol) for arg in args]):
            return Expression(op, args)
        else:
            # then all the args are floats: just return a float!
            realized_fn = _lookup_operator(op.name)
            return realized_fn(*args)

    # define sin, etc., in the parameter and expression so that np.sin(a)
    # will resolve to Expression('sin', tuple(a)), etc.
    setattr(OperatorMixin, fn_name, fn)
    # The np.sin(a) trick only works for a limited set of functions
    # defined by numpy itself. For arbitrary user defined functions
    # we add them to the bumps.pmath namespace so the user can find them.
    setattr(pmath, fn_name, fn)


def _build_operator_mixin():
    unary_op = set(("pos", "neg", "abs"))
    binary_op = set(("add", "sub", "mul", "floordiv", "truediv", "pow"))
    math_fn = set(v.name for v in Operators) - unary_op - binary_op
    for op_name in unary_op:
        _make_unary_op(op_name)
    for op_name in binary_op:
        _make_binary_op(op_name)
    # By adding the math functions to the mixin, calling np.sin(parameter) or
    # np.sin(expression) will return the generated expression for the object.
    for fn_name in math_fn:
        _make_math_fn(fn_name)


_build_operator_mixin()

UserFunctionRegistry: Dict[str, Callable[..., float]] = {}


# TODO: allow schema validation on user-defined functions
@dataclass(init=False)
class UserFunction:
    """
    User-defined functions.

    This is a helper class for the @function decorator, which treats the
    operator as one of the possible expression operators.

    These won't be properly serialized/deserialized through the JSON schema
    unless the function is registered in advance. The schema will not include
    these functions as possible values even if registered, so a schema
    validator may fail on one of these functions.
    """

    name: str

    # A function registry to remember the code associated with the name.
    # This is a class attribute, so it is initialized with an empty dict().
    # Ignore complaints from lint.
    # TODO: use pmath as our registry of available functions.
    def __init__(self, fn: Callable):
        name = fn.__name__
        if name in UserFunctionRegistry:
            raise TypeError(f"Function {name} already registered in bumps.")
        UserFunctionRegistry[name] = fn
        self.name = name


def function(fn: Callable):
    """
    Convert a function into a delayed evaluator.

    The value of the function is computed from the values of the parameters
    at the time that the function value is requested rather than when the
    function is created.
    """
    name = fn.__name__
    op = UserFunction(fn)

    def wrapped(*args: "ValueType"):
        return Expression(op, args)

    wrapped.__name__ = fn.__name__
    wrapped.__doc__ = fn.__doc__ if fn.__name__.endswith("d") else f"{fn.__name__}(Parameter)"
    # Add the symbol to pmath
    setattr(pmath, name, wrapped)
    pmath.__all__.append(name)
    return wrapped


# min/max
min = function(builtins.min)
max = function(builtins.max)


# Trig functions defined in degrees rather than radians.
@function
def cosd(v):
    """Return the cosine of x (measured in in degrees)."""
    return np.cos(np.radians(v))


@function
def sind(v):
    """Return the sine of x (measured in in degrees)."""
    return np.sin(np.radians(v))


@function
def tand(v):
    """Return the tangent of x (measured in in degrees)."""
    return np.tan(np.radians(v))


@function
def arccosd(v):
    """Return the arc cosine (measured in in degrees) of x."""
    return np.degrees(np.arccos(v))


@function
def arcsind(v):
    """Return the arc sine (measured in in degrees) of x."""
    return np.degrees(np.arcsin(v))


@function
def arctand(v):
    """Return the arc tangent (measured in in degrees) of x."""
    return np.degrees(np.arctan(v))


@function
def arctan2d(dy, dx):
    """Return the arc tangent (measured in in degrees) of y/x.
    Unlike atan(y/x), the signs of both x and y are considered."""
    return np.degrees(np.arctan2(dy, dx))


# Aliases for arcsin, etc., both here in bumps.parameters and in bumps.pmath.
pmath.asin = asin = pmath.arcsin
pmath.acos = acos = pmath.arccos
pmath.atan = atan = pmath.arctan
pmath.atan2 = atan2 = pmath.arctan2

pmath.asind = asind = arcsind
pmath.acosd = acosd = arccosd
pmath.atand = atand = arctand
pmath.atan2d = atan2d = arctan2d

pmath.asinh = asinh = pmath.arcsinh
pmath.acosh = acosh = pmath.arccosh
pmath.atanh = atanh = pmath.arctanh

pmath.__all__.extend(
    (
        "asin",
        "acos",
        "atan",
        "atan2",
        "asind",
        "acosd",
        "atand",
        "atan2d",
        "asinh",
        "acosh",
        "atanh",
    )
)

# restate these for export, now that they're all defined:
ValueType = Union[Parameter, Expression, Calculation, float]


@dataclass(init=False)
class ParameterSet:
    """
    A parameter that depends on the model.
    """

    names: Optional[List[str]]
    reference: Parameter
    parameterlist: Optional[List[Parameter]]

    def __init__(
        self, reference: Parameter, names: Optional[List[str]] = None, parameterlist: Optional[List[Parameter]] = None
    ):
        """
        Create a parameter set, with one parameter for each model name.

        *names* is the list of model names.

        *reference* is the underlying :class:`parameter.Parameter` that will
        be set when the model is selected.

        *parameters* will be created, with one parameter per model.
        """
        names = names if names is not None else []
        self.names = names
        self.reference = reference
        # TODO: explain better why parameters are using np.array
        # Force numpy semantics on slice operations by using an array
        # of objects rather than a list of objects
        if parameterlist is not None:
            # we are being reinitialized with parameters
            self.parameters = np.array(parameterlist)
        else:
            self.parameters = np.array([copy(reference) for _ in names])
        # print self.reference, self.parameters
        for p, n in zip(self.parameters, names):
            p.name = " ".join((n, p.name))

        # N.B. if the reference parameter is not referenced anywhere in the models,
        # it will no longer show up in FitProblem.parameters
        # self.__class__.parameterlist = property(self._get_parameterlist) #lambda self: self.parameters.tolist())

    @property
    def parameterlist(self) -> List[Parameter]:
        return self.parameters.tolist()

    def to_dict(self):
        return {
            "type": "ParameterSet",
            "names": self.names,
            "reference": to_dict(self.reference),
            # Note: parameters are stored in a numpy array
            "parameters": to_dict(self.parameters.tolist()),
        }

    # Make the parameter set act like a list
    def __getitem__(self, i):
        """
        Return the underlying parameter for the model index.  Index can
        either be an integer or a model name.  It can also be a slice,
        in which case a new parameter set is returned.
        """
        # Try looking up the free variable by model name rather than model
        # index. If this fails, assume index is a model index.
        try:
            i = self.names.index(i)
        except ValueError:
            pass
        if isinstance(i, slice):
            obj = copy(self)
            obj.names = self.names[i]
            obj.reference = self.reference
            obj.parameters = self.parameters[i]
            return obj
        return self.parameters[i]

    def __setitem__(self, i, v):
        """
        Set the underlying parameter for the model index.  Index can
        either be an integer or a model name.  It can also be a slice,
        in which case all underlying parameters are set, either to the
        same value if *v* is a single parameter, otherwise *v* must have
        the same length as the slice.
        """
        try:
            i = self.names.index(i)
        except ValueError:
            pass
        self.parameters[i] = v

    def __iter__(self):
        return iter(self.parameters)

    def __len__(self):
        return len(self.parameters)

    def set_model(self, index):
        """
        Set the underlying model parameter to the value of the nth model.
        """
        self.reference.value = self.parameters[index].value

    def get_model(self, index):
        """
        Get the reference and underlying model parameter for the nth model.
        """
        return (id(self.reference), self.parameters[index])

    @property
    def values(self):
        return [p.value for p in self.parameters]

    @values.setter
    def values(self, values):
        for p, v in zip(self.parameters, values):
            p.value = v

    def range(self, *args, **kw):
        """
        Like :meth:`Parameter.range`, but applied to all models.
        """
        for p in self.parameters:
            p.range(*args, **kw)

    def pm(self, *args, **kw):
        """
        Like :meth:`Parameter.pm`, but applied to all models.
        """
        for p in self.parameters:
            p.pm(*args, **kw)

    def pmp(self, *args, **kw):
        """
        Like :meth:`Parameter.pmp`, but applied to all models.
        """
        for p in self.parameters:
            p.pmp(*args, **kw)


class Reference(Parameter):
    """
    Create an adaptor so that a model attribute can be treated as if it
    were a parameter.  This allows only direct access, wherein the
    storage for the parameter value is provided by the underlying model.

    Indirect access, wherein the storage is provided by the parameter, cannot
    be supported since the parameter has no way to detect that the model
    is asking for the value of the attribute.  This means that model
    attributes cannot be assigned to parameter expressions without some
    trigger to update the values of the attributes in the model.

    NOTE: this class can not be serialized with a dataclass schema
    TODO: can sasmodels just use Parameter directly?
    """

    def __init__(self, obj, attr, **kw):
        self.obj = obj
        self.attr = attr
        kw.setdefault("name", ".".join([obj.__class__.__name__, attr]))
        Parameter.__init__(self, **kw)

    @property
    def value(self):
        return getattr(self.obj, self.attr)

    @value.setter
    def value(self, value):
        setattr(self.obj, self.attr, value)


@dataclass(init=False)
class FreeVariables(object):
    """
    A collection of parameter sets for a group of models.

    *names* is the set of model names.

    The parameters themselves are specified as key=value pairs, with key
    being the attribute name which is used to retrieve the parameter set
    and value being a :class:`Parameter` containing the parameter that is
    shared between the models.

    In order to evaluate the log likelihood of all models simultaneously,
    the fitting program will need to call set_model with the model index
    for each model in turn in order to substitute the values from the free
    variables into the model.  This allows us to share a common sample
    across multiple data sets, with each dataset having its own values for
    some of the sample parameters.  The alternative is to copy the entire
    sample structure, sharing references to common parameters and creating
    new parameters for each model for the free parameters.  Setting up
    these copies was inconvenient.
    """

    names: List[str]
    parametersets: Dict[str, ParameterSet]

    def __init__(self, names=None, parametersets=None, **kw):
        if names is None:
            raise TypeError("FreeVariables needs name=[model1, model2, ...]")
        self.names = names
        if parametersets is not None:
            # assume that we are initializing with a dict of
            # fully initialized ParameterSet objects
            self.parametersets = parametersets
        else:
            # we are initializing with kw = Dict[key, (list of Parameters)]
            # Create slots to hold the free variables
            self.parametersets = dict((k, ParameterSet(v, names=names)) for k, v in kw.items())

    # Shouldn't need explicit __getstate__/__setstate__ but mpi4py pickle
    # chokes without it.
    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state

    def __getattr__(self, k):
        """
        Return the parameter set for the given free parameter.
        """
        try:
            return self.parametersets[k]
        except KeyError:
            raise AttributeError("FreeVariables has no attribute %r" % k)

    def parameters(self):
        """
        Return the set of free variables for all the models.
        """
        return dict((k, v.parameters) for k, v in self.parametersets.items())

    def to_dict(self):
        return {"type": type(self).__name__, "names": self.names, "parameters": to_dict(self.parametersets)}

    def set_model(self, i):
        """
        Set the reference parameters for model *i*.
        """
        for p in self.parametersets.values():
            p.set_model(i)

    def get_model(self, i):
        """
        Get the parameters for model *i* as {reference: substitution}
        """
        return dict(p.get_model(i) for p in self.parametersets.values())


def flatten(s):
    if isinstance(s, (tuple, list, np.ndarray)):
        return reduce(lambda a, b: a + flatten(b), s, [])
    elif isinstance(s, set):
        raise TypeError("parameter flattening cannot order sets")
    elif isinstance(s, dict):
        return reduce(lambda a, b: a + flatten(s[b]), sorted(s.keys()), [])
    elif isinstance(s, ValueProtocol):
        return [s]
    elif s is None:
        return []
    else:
        raise TypeError("don't understand type %s for %r" % (type(s), s))


def format(p, indent=0, freevars=None, field=None):
    """
    Format parameter set for printing.

    Note that this only says how the parameters are arranged, not how they
    relate to each other.
    """
    freevars = {} if freevars is None else freevars
    p = freevars.get(id(p), p)
    if isinstance(p, dict) and p != {}:
        res = []
        for k in sorted(p.keys()):
            if k.startswith("_"):
                continue
            s = format(p[k], indent + 2, field=k, freevars=freevars)
            label = " " * indent + "." + k
            if s.endswith("\n"):
                res.append(label + "\n" + s)
            else:
                res.append(label + " = " + s + "\n")
        if "_index" in p:
            res.append(format(p["_index"], indent, freevars=freevars))
        return "".join(res)

    elif isinstance(p, (list, tuple, np.ndarray)) and len(p):
        res = []
        for k, v in enumerate(p):
            s = format(v, indent + 2, freevars=freevars)
            label = " " * indent + "[%d]" % k
            if s.endswith("\n"):
                res.append(label + "\n" + s)
            else:
                res.append(label + " = " + s + "\n")
        return "".join(res)

    elif isinstance(p, Parameter):
        s = ""
        if str(p) != field:
            s += str(p) + " = "
        s += "%g" % p.value
        if not p.fixed:
            if p.prior is not None:
                bounds = p.prior.limits
            elif p.bounds is not None:
                bounds = p.bounds
            else:
                bounds = p.limits
            s += " in [%g,%g]" % tuple(bounds)
        return s

    elif isinstance(p, Parameter):
        return "%s = %g" % (str(p), p.value)

    else:
        return str(p)


def summarize(pars, sorted=False):
    """
    Return a stylized list of parameter names and values with range bars
    suitable for printing.

    If sorted, then print the parameters sorted alphabetically by name.
    """
    output = []
    if sorted:
        pars = sorted(pars, key=lambda x: x.name)
    for p in pars:
        if not isfinite(p.value):
            bar = ["*invalid* "]
        else:
            position = int(p.prior.get01(p.value) * 9.999999999)
            bar = ["."] * 10
            if position < 0:
                bar[0] = "<"
            elif position > 9:
                bar[9] = ">"
            else:
                bar[position] = "|"
        output.append("%40s %s %10g in %s" % (p.name, "".join(bar), p.value, p.bounds))
    return "\n".join(output)


def unique(s) -> List[Parameter]:
    """
    Return the unique set of parameters

    The ordering is stable.  The same parameters/dependencies will always
    return the same ordering, with the first occurrence first.
    """
    # Walk structures such as dicts and lists
    pars = flatten(s)
    # print "====== flattened"
    # print "\n".join("%s:%s"%(id(p),p) for p in pars)
    # Also walk parameter expressions
    pars = pars + flatten([p.parameters() for p in pars])
    # print "====== extended"
    # print "\n".join("%s:%s"%(id(p),p) for p in pars)

    # TODO: implement n log n rather than n^2 uniqueness algorithm
    # problem is that the sorting has to be unique across a pickle.
    result = []
    for p in pars:
        if not any(p is q for q in result):
            result.append(p)

    # print "====== unique"
    # print "\n".join("%s:%s"%(id(p),p) for p in result)
    # Return the complete set of parameters
    return result


def fittable(s):
    """
    Return the list of fittable parameters in no paraticular order.

    Note that some fittable parameters may be fixed during the fit.
    """
    return [p for p in unique(s) if p.fittable]


def varying(s: List[Parameter]) -> List[Parameter]:
    """
    Return the list of fitted parameters in the model.

    This is the set of parameters that will vary during the fit.
    """
    return [p for p in unique(s) if not p.fixed]


def _has_prior(p: Parameter) -> bool:
    prior = getattr(p, "prior", None)
    limits = getattr(prior, "limits", (-np.inf, np.inf))
    return prior is not None and not isinstance(prior, mbounds.Unbounded) and limits != (-np.inf, np.inf)


def priors(s: List[Parameter]) -> List[Parameter]:
    """
    Return the list of parameters (fitted or computed) that have prior
    probabilities associated with them. This includes all varying parameters,
    plus expressions (including simple links), but ignoring constants and
    fixed parameters whose probabilities won't change the fits.
    """
    return [p for p in unique(s) if _has_prior(p)]


def randomize(s: List[Parameter]):
    """
    Set random values to the parameters in the parameter set, with
    values chosen according to the bounds.
    """
    for p in s:
        p.value = p.prior.random(1)[0]


def current(s: List[Parameter]):
    return [p.value for p in s]


# ========= trash ===================


def copy_linked(has_parameters, free_names=None):
    """
    make a copy of an object with parameters
     - then link all the parameters, except
     - those with names matching "free_names"
    """
    assert callable(getattr(has_parameters, "parameters", None)) == True
    from copy import deepcopy

    copied = deepcopy(has_parameters)
    free_names = [] if free_names is None else free_names
    original_pars = unique(has_parameters.parameters())
    copied_pars = unique(copied.parameters())
    for op, cp in zip(original_pars, copied_pars):
        if not op.name in free_names:
            cp.slot = op.slot
        else:
            cp.id = str(uuid.uuid4())
    return copied


# ==== Comparison operators ===
class Comparisons(Enum):
    """comparison operators"""

    gt = ">"
    ge = ">="
    le = "<="
    lt = "<"
    # eq = '=='
    # ne = '!='


@dataclass(init=False)
class Constraint:
    """Express inequality constraints between model elements"""

    fixed = True

    op: Comparisons
    a: ValueType
    b: ValueType

    def __init__(self, a, b, op):
        import operator

        object.__setattr__(self, "a", a)
        object.__setattr__(self, "b", b)
        op_name = str(Comparisons(op).name)
        object.__setattr__(self, "compare", getattr(operator, op_name.lower()))
        object.__setattr__(self, "op", op)

    # TODO: is this really necessary?  What is the reason for this trap?
    # It seems like being able to cast with bool(Constraint) would be
    # useful in some circumstances, like doing max(List[Parameter]), which
    # currently fails.
    def __bool__(self):
        raise TypeError("failed bool")

    __nonzero__ = __bool__

    def __float__(self):
        """return a float value that can be differentiated"""
        return 0.0 if self.satisfied else abs(float(self.a) - float(self.b))

    def __str__(self):
        return "(%s %s %s)" % (self.a, self.op, self.b)

    @property
    def satisfied(self):
        return self.compare(float(self.a), float(self.b))


def _make_constraint(op_str: str) -> Callable[..., Constraint]:
    return lambda self, other: Constraint(self, other, op_str)


def _build_constraints_mixin():
    for comp_item in Comparisons:
        op_name = comp_item.name
        op_str = comp_item.value
        setattr(OperatorMixin, f"__{op_name}__", _make_constraint(op_str))


_build_constraints_mixin()


class Alias(object):
    """
    Parameter alias.

    Rather than modifying a model to contain a parameter slot,
    allow the parameter to exist outside the model. The resulting
    parameter will have the full parameter semantics, including
    the ability to replace a fixed value with a parameter expression.

    """

    def __init__(self, obj, attr, p=None, name=None):
        self.obj = obj
        self.attr = attr
        if name is None:
            name = ".".join([obj.__class__.__name__, attr])
        self.p = Parameter.default(p, name=name)

    def update(self):
        setattr(self.obj, self.attr, self.p.value)

    def parameters(self):
        return self.p.parameters()

    def to_dict(self):
        return {
            "type": type(self).__name__,
            "p": to_dict(self.p),
            # TODO: can't json pickle arbitrary objects
            "obj": to_dict(self.obj),
            "attr": self.attr,
        }


def substitute(a):
    """
    Return structure a with values substituted for all parameters.

    The function traverses lists, tuples and dicts recursively.  Things
    which are not parameters are returned directly.
    """
    if isinstance(a, ValueProtocol):
        return float(a.value)
    elif isinstance(a, tuple):
        return tuple(substitute(v) for v in a)
    elif isinstance(a, list):
        return [substitute(v) for v in a]
    elif isinstance(a, dict):
        return dict((k, substitute(v)) for k, v in a.items())
    elif isinstance(a, np.ndarray):
        return np.array([substitute(v) for v in a])
    else:
        return a


class Function(ValueProtocol):
    """
    **DEPRECATED**

    Delayed function evaluator.

    f.value evaluates the function with the values of the
    parameter arguments at the time f.value is referenced rather
    than when the function was invoked.
    """

    __slots__ = ["op", "args", "kw"]
    op: Callable[..., float]
    args: Optional[Any]
    kw: Dict[Any, Any]

    def __init__(self, op, *args, **kw):
        warnings.warn("Function no longer supported", DeprecationWarning, stacklevel=1)
        self.name = kw.pop("name", None)
        self.op, self.args, self.kw = op, args, kw
        self._parameters = self._find_parameters()

    def _find_parameters(self):
        # Figure out which arguments to the function are parameters
        # deps = [p for p in self.args if isinstance(p, ValueProtocol)]
        args = [arg for arg in self.args if isinstance(arg, ValueProtocol)]
        kw = dict((name, arg) for name, arg in self.kw.items() if isinstance(arg, ValueProtocol))
        deps = flatten((args, kw))
        # Find out which other parameters these parameters depend on.
        res = []
        for p in deps:
            res.extend(p.parameters())
        return res

    def parameters(self):
        return self._parameters

    def _value(self):
        # Expand args and kw, replacing instances of parameters
        # with their values
        return self.op(*substitute(self.args), **substitute(self.kw))

    value = property(_value)

    def to_dict(self):
        return {
            "type": "Function",
            "name": self.name,
            # TODO: function not stored properly in json
            "op": to_dict(self.op),
            "args": to_dict(self.args),
            "kw": to_dict(self.kw),
        }

    def __getstate__(self):
        return self.name, self.op, self.args, self.kw

    def __setstate__(self, state):
        self.name, self.op, self.args, self.kw = state
        self._parameters = self._find_parameters()

    def __str__(self):
        if self.name is not None:
            name = self.name
        else:
            args = [str(v) for v in self.args]
            kw = [str(k) + "=" + str(v) for k, v in self.kw.items()]
            name = self.op.__name__ + "(" + ", ".join(args + kw) + ")"
        return name
        # return "%s:%g" % (name, self.value)


# ===== Tests ====


def test_operator():
    a = Parameter(1, name="a")
    b = Parameter(2, name="b")
    c = Parameter(3, name="c")
    C = Constant(5, name="C")

    assert a.fixed

    # Check strings
    assert str(a + b) == "a + b"
    assert (a + b).name == "a + b"
    assert str(-a) == "-a"
    assert (-a).value == -a.value
    assert str(a + b * c) == "a + b * c"
    assert str((a + b) * c) == "(a + b) * c"
    assert str(np.sin(a + b) * c) == "sin(a + b) * c"
    assert str(a + C) == "a + C"
    assert str(a + C + 3) == "a + C + 3"
    assert str(3 + a + C) == "3 + a + C"
    assert str(a.sin()) == "sin(a)"
    assert str(atan2(a, b)) == "arctan2(a, b)"
    # float(expr) evaluates the expression; it doesn't build an expr with float.

    # Check parameters
    assert (a + b).parameters() == [a, b]
    assert (np.sin(a + b) * c).parameters() == [a, b, c]

    # Check values
    a.value = 3
    assert (a + b).value == 5.0
    assert float(a + b) == a.value + b.value
    assert a.sin().value == np.sin(a.value)
    assert (3 + a + C).value == 3 + 3 + 5
    assert np.sin(a + b).value == np.sin(a.value + b.value)
    assert atan2(a, b).value == atan2(a.value, b.value)

    # Make sure that evaluation is lazy. Capture the expression with one
    # set of values for the parameters, update them with a new set of values,
    # then check if the result is what you get when you call the function
    # directly on those new values.
    scope = locals()  # record the currently available parameter handles

    def capture_test(expr, result, **kw):
        # print("checking", expr, "for", kw, "yields", result)
        saved = {k: scope[k].value for k in kw}
        for k, v in kw.items():
            scope[k].value = float(v)
        try:
            assert expr.value == result, f"for {expr} expected {result} but got {expr.value}"
        finally:
            for k, v in saved.items():
                scope[k].value = v

    capture_test(np.sin(a + b), np.sin(0.5 + 3), a=0.5, b=3)
    capture_test(np.arctan2(a, b), atan2(0.5, 3), a=0.5, b=3)
    capture_test(np.round(a), np.round(-0.6), a=-0.6)
    capture_test(min(a, b), builtins.min(-0.6, 3), a=-0.6, b=3)
    capture_test(min(a, b, -2), builtins.min(-0.6, 3, -2), a=-0.6, b=3)
    capture_test(abs(a), 2.5, a=-2.5)

    # Check that symbols are defined in pmath
    capture_test(pmath.sind(a), np.sin(np.radians(25)), a=25)
    assert "sind" in pmath.__all__

    # TODO: can we evaluate an expression for an entire population at once?

    # Check slots
    limited = Parameter(3, name="limited", limits=[0.5, 1.5], bounds=[0, 1])
    limited.add_prior()
    assert np.isinf(limited.nllf())
    assert np.isinf(limited.nllf())
    limited.value = 0.6
    assert limited.nllf() == 0.0
    limited.value = 0.2
    assert np.isinf(limited.nllf())

    limited.equals(a + b)
    assert limited.value == (a + b).value

    assert np.isinf(limited.nllf())
    a.value = b.value = 0.1
    assert np.isinf(limited.nllf())
    a.value = b.value = 0.3
    assert limited.nllf() == 0.0
    try:
        limited.value = 5
        failed = True
    except Exception:
        # TODO: define which error improper assignment should raise
        # Currently this raises an attribute error on limited.slot.value
        failed = False
    if failed:
        raise RuntimeError("failed to raise error when assigning value to expression")

    # Check parameter list operations
    s = [a, limited]
    assert unique(s) == [a, limited, b]
    assert fittable(s) == [a, b]
    assert varying(s) == []
    b.range(0, 3)
    assert not b.fixed
    assert varying(s) == [b]
    assert current(s) == [a.value, limited.value]

    # Check normal deviation
    mu, sigma = 3, 2
    b.dev(sigma, mean=mu)
    b.value = 4
    b.add_prior()
    nllf_target = 0.5 * ((b.value - mu) / sigma) ** 2 + np.log(2 * np.pi * sigma**2) / 2
    assert abs(b.nllf() - nllf_target) / nllf_target < 1e-12

    # Check conditions
    a.value, b.value = 3, 4
    capture = a < b
    assert isinstance(capture, Constraint)
    assert capture.satisfied
    a.value, b.value = 4, 3
    assert not capture.satisfied

    scope = locals()

    def raises(condition_str, exception):
        try:
            eval(condition_str, locals=scope)
        except exception:
            pass
        else:
            raise AssertionError(f"{condition_str} does not raise {exception}")

    raises("a < b < c", TypeError)
    raises("a < b and b < c", TypeError)
    raises("a < b or b < c", TypeError)
    raises("not (a < b)", TypeError)
    raises("not a", TypeError)
    raises("a and b", TypeError)
    raises("a or b", TypeError)


if __name__ == "__main__":
    test_operator()
